#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
SR back-end package for Kaldi
============================================================================

"""

import logging
import struct
import sys

_log = logging.getLogger("engine.kaldi")


#---------------------------------------------------------------------------

# Module level singleton instance of this engine implementation.
_engine = None


def is_engine_available(**kwargs):
    """ Check whether Kaldi is available. """
    global _engine
    if _engine:
        return True

    if struct.calcsize("P") == 4:  # 32-bit
        _log.warning("The Python environment is 32-bit.  Kaldi requires a "
                     "64-bit python environment.")
        return False

    if sys.version_info.major < 3 or sys.version_info.minor < 6:
        _log.warning("This version of Python is not compatible with the "
                     "Kaldi back-end.  Python version 3.6 or higher is "
                     "required.")
        return False

    # Attempt to import the engine class from the module.
    try:
        from .engine import KaldiEngine
        return True
    except ImportError as e:
        _log.warning("Failed to import from Kaldi engine module: %s", e)
        return False
    except Exception as e:
        _log.exception("Exception during import of Kaldi engine module: %s",
                       e)
        return False

    return True


def get_engine(**kwargs):
    """ Retrieve the Kaldi back-end engine object. """
    global _engine
    if not _engine:
        from .engine import KaldiEngine
        _engine = KaldiEngine(**kwargs)
    return _engine



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Number language element class
============================================================================

"""

from ...grammar.elements  import Alternative, Repetition, Sequence, RuleWrap
from ..loader             import language
from .integer             import Integer


#---------------------------------------------------------------------------
# Number class.

class Number(Alternative):

    _int_max = 1000000
    _ser_len = 8

    def __init__(self, name=None, zero=False, default=None):
        name = str(name)
        int_name = "_Number_int_" + name
        if zero:  int_min = 0
        else:     int_min = 1
        single = Integer(None, int_min, self._int_max)

        ser_name = "_Number_ser_" + name
        item = Integer(None, 0, 100)
        if zero:
            series = Repetition(item, 1, self._ser_len)
        else:
            first = Integer(None, 1, 100)
            repetition = Repetition(item, 0, self._ser_len - 1)
            series = Sequence([first, repetition])

        children = [single, series]
        Alternative.__init__(self, children, name=name, default=default)

    def value(self, node):
        value = Alternative.value(self, node)

        if isinstance(value, list):
            items = []
            for item in value:
                if isinstance(item, list):
                    items.extend(item)
                else:
                    items.append(item)
            value = 0
            for item in items:
                if item < 10:  factor = 10
                else:          factor = 100
                value *= factor
                value += item

        return value


#---------------------------------------------------------------------------
# Number reference class.

class NumberRef(RuleWrap):

    def __init__(self, name=None, zero=False, default=None):
        element = Number(None, zero=zero)
        RuleWrap.__init__(self, name, element, default=default)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements an interface to the system clipboard using pyperclip
and should work on Windows, Mac OS and Linux-based operating systems.
"""

# pylint: disable=W0622
# Suppress warnings about redefining the built-in 'format' function.

import os

from six import integer_types
import pyperclip

from .base_clipboard import BaseClipboard


class PyperclipClipboard(BaseClipboard):
    """
    Class for interacting with the system clipboard via the
    `pyperclip <https://pyperclip.readthedocs.io/en/latest/>`__ Python
    package.

    This is Dragonfly's default clipboard class on platforms other than
    Windows.

    .. note::

       This class does work on Windows, however the Windows
       :class:`dragonfly.windows.win32_clipboard.Win32Clipboard` class
       should be used instead because this class doesn't support as many
       clipboard formats.

    """

    #-----------------------------------------------------------------------

    @classmethod
    def get_system_text(cls):
        return pyperclip.paste()

    @classmethod
    def set_system_text(cls, content):
        # If *content* is None, clear the clipboard and return early.
        if content is None:
            cls.clear_clipboard()
            return

        # Otherwise, convert *content* and set the system clipboard data.
        content = cls.convert_format_content(cls.format_unicode, content)
        pyperclip.copy(content)

    @classmethod
    def clear_clipboard(cls):
        cls.set_system_text(u"")

    # ----------------------------------------------------------------------

    def copy_from_system(self, formats=None, clear=False):
        # Determine the supported formats.  Copied file paths may be
        #  available to us on X11.
        supported_formats = [self.format_text, self.format_unicode]
        if os.environ.get("XDG_SESSION_TYPE") == "x11":
            supported_formats.append(self.format_hdrop)

        # Determine which formats to retrieve.
        caller_specified_formats = bool(formats)
        if not formats:
            formats = supported_formats
        elif isinstance(formats, integer_types):
            formats = (formats,)

        # Verify that the given formats are valid.
        for format in formats:
            if not isinstance(format, integer_types):
                raise TypeError("Invalid clipboard format: %r"
                                % format)

        # Retrieve the system clipboard content.
        # This class only handles text formats.
        text = self.get_system_text()
        contents = {}

        # Populate the clipboard contents dictionary and raise errors for
        #  unavailable formats.
        for format in formats:
            try:
                err = False
                content = self.convert_format_content(format, text)
                contents[format] = content
            except (ValueError, TypeError):
                err = True

            # Do not raise an error for CF_HDROP content if the caller did
            #  not specify it.
            if format == self.format_hdrop and not caller_specified_formats:
                err = False

            # Always raise an error if the format is not supported.
            if format not in supported_formats:
                err = True

            # Raise an error if a specified clipboard format was not
            #  available.
            if err:
                format_repr = self.format_names.get(format, format)
                message = ("Specified clipboard format %r is not "
                           "available." % format_repr)
                raise TypeError(message)
        self._contents = contents

        # Then clear the system clipboard, if requested.
        if clear:
            self.clear_clipboard()

    def copy_to_system(self, clear=True):
        # Clear the system clipboard, if requested.
        if clear:
            self.clear_clipboard()

        # Transfer text content to system clipboard, if necessary.
        text = self.get_text()
        if text is not None:
            self.set_system_text(text)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This script generates code for in the typeables.py and action_key.py
    files.

"""


(vkey, lookup) = range(2)

key_map = (
    ("Lowercase letter keys", 0, [
        (lookup, "a",                              "a alpha"),
        (lookup, "b",                              "b bravo"),
        (lookup, "c",                              "c charlie"),
        (lookup, "d",                              "d delta"),
        (lookup, "e",                              "e echo"),
        (lookup, "f",                              "f foxtrot"),
        (lookup, "g",                              "g golf"),
        (lookup, "h",                              "h hotel"),
        (lookup, "i",                              "i india"),
        (lookup, "j",                              "j juliet"),
        (lookup, "k",                              "k kilo"),
        (lookup, "l",                              "l lima"),
        (lookup, "m",                              "m mike"),
        (lookup, "n",                              "n november"),
        (lookup, "o",                              "o oscar"),
        (lookup, "p",                              "p papa"),
        (lookup, "q",                              "q quebec"),
        (lookup, "r",                              "r romeo"),
        (lookup, "s",                              "s sierra"),
        (lookup, "t",                              "t tango"),
        (lookup, "u",                              "u uniform"),
        (lookup, "v",                              "v victor"),
        (lookup, "w",                              "w whisky"),
        (lookup, "x",                              "x xray"),
        (lookup, "y",                              "y yankee"),
        (lookup, "z",                              "z zulu")]),
    ("Uppercase letter keys", 0, [
        (lookup, "A",                              "A Alpha"),
        (lookup, "B",                              "B Bravo"),
        (lookup, "C",                              "C Charlie"),
        (lookup, "D",                              "D Delta"),
        (lookup, "E",                              "E Echo"),
        (lookup, "F",                              "F Foxtrot"),
        (lookup, "G",                              "G Golf"),
        (lookup, "H",                              "H Hotel"),
        (lookup, "I",                              "I India"),
        (lookup, "J",                              "J Juliet"),
        (lookup, "K",                              "K Kilo"),
        (lookup, "L",                              "L Lima"),
        (lookup, "M",                              "M Mike"),
        (lookup, "N",                              "N November"),
        (lookup, "O",                              "O Oscar"),
        (lookup, "P",                              "P Papa"),
        (lookup, "Q",                              "Q Quebec"),
        (lookup, "R",                              "R Romeo"),
        (lookup, "S",                              "S Sierra"),
        (lookup, "T",                              "T Tango"),
        (lookup, "U",                              "U Uniform"),
        (lookup, "V",                              "V Victor"),
        (lookup, "W",                              "W Whisky"),
        (lookup, "X",                              "X Xray"),
        (lookup, "Y",                              "Y Yankee"),
        (lookup, "Z",                              "Z Zulu")]),
    ("Number keys", 0, [
        (lookup, "0",                              "0 zero"),
        (lookup, "1",                              "1 one"),
        (lookup, "2",                              "2 two"),
        (lookup, "3",                              "3 three"),
        (lookup, "4",                              "4 four"),
        (lookup, "5",                              "5 five"),
        (lookup, "6",                              "6 six"),
        (lookup, "7",                              "7 seven"),
        (lookup, "8",                              "8 eight"),
        (lookup, "9",                              "9 nine")]),
    ("Symbol keys", 0, [
        (lookup, "!",                              "! bang exclamation"),
        (lookup, "@",                              "@ at"),
        (lookup, "#",                              "# hash"),
        (lookup, "$",                              "$ dollar"),
        (lookup, "%",                              "% percent"),
        (lookup, "^",                              "^ caret"),
        (lookup, "&",                              "& and ampersand"),
        (lookup, "*",                              "* star asterisk"),
        (lookup, "(",                              "( leftparen lparen"),
        (lookup, ")",                              ") rightparen rparen"),
        (lookup, "-",                              "- minus hyphen"),
        (lookup, "_",                              "_ underscore"),
        (lookup, "+",                              "+ plus"),
        (lookup, "`",                              "` backtick"),
        (lookup, "~",                              "~ tilde"),
        (lookup, "[",                              "[ leftbracket lbracket"),
        (lookup, "]",                              "] rightbracket rbracket"),
        (lookup, "{",                              "{ leftbrace lbrace"),
        (lookup, "}",                              "} rightbrace rbrace"),
        (lookup, "\\\\",                           "\\\\ backslash"),
        (lookup, "|",                              "| bar"),
        (lookup, ":",                              ": colon"),
        (lookup, ";",                              "; semicolon"),
        (lookup, "\\\'",                           "\\\' apostrophe singlequote squote"),
        (lookup, '\\\"',                           '\\\" quote doublequote dquote'),
        (lookup, ",",                              ", comma"),
        (lookup, ".",                              ". dot"),
        (lookup, "/",                              "/ slash"),
        (lookup, "<",                              "< lessthan leftangle langle"),
        (lookup, ">",                              "> greaterthan rightangle rangle"),
        (lookup, "?",                              "? question"),
        (lookup, "=",                              "= equal equals")]),
    ("Whitespace and editing keys", 1, [
        (vkey,   "key_symbols.RETURN",             "enter"),
        (vkey,   "key_symbols.TAB",                "tab"),
        (vkey,   "key_symbols.SPACE",              "space"),
        (vkey,   "key_symbols.BACK",               "backspace"),
        (vkey,   "key_symbols.DELETE",             "delete del")]),
    ("Main modifier keys", 1, [
        (vkey,   "key_symbols.SHIFT",              "shift"),
        (vkey,   "key_symbols.CONTROL",            "control ctrl"),
        (vkey,   "key_symbols.ALT",                "alt")]),
    ("Right modifier keys", 1, [
        (vkey,   "key_symbols.RSHIFT",             "rshift"),
        (vkey,   "key_symbols.RCONTROL",           "rcontrol rctrl"),
        (vkey,   "key_symbols.RALT",               "ralt")]),
    ("Special keys", 1, [
        (vkey,   "key_symbols.ESCAPE",             "escape"),
        (vkey,   "key_symbols.INSERT",             "insert"),
        (vkey,   "key_symbols.PAUSE",              "pause"),
        (vkey,   "key_symbols.LSUPER",             "win"),
        (vkey,   "key_symbols.RSUPER",             "rwin"),
        (vkey,   "key_symbols.APPS",               "apps popup"),
        (vkey,   "key_symbols.SNAPSHOT",           "snapshot printscreen")]),
    ("Lock keys", 1, [
        (vkey,   "key_symbols.SCROLL_LOCK",        "scrolllock"),
        (vkey,   "key_symbols.NUM_LOCK",           "numlock"),
        (vkey,   "key_symbols.CAPS_LOCK",          "capslock")]),
    ("Navigation keys", 1, [
        (vkey,   "key_symbols.UP",                 "up"),
        (vkey,   "key_symbols.DOWN",               "down"),
        (vkey,   "key_symbols.LEFT",               "left"),
        (vkey,   "key_symbols.RIGHT",              "right"),
        (vkey,   "key_symbols.PAGE_UP",            "pageup pgup"),
        (vkey,   "key_symbols.PAGE_DOWN",          "pagedown pgdown"),
        (vkey,   "key_symbols.HOME",               "home"),
        (vkey,   "key_symbols.END",                "end")]),
    ("Number pad keys", 1, [
        (vkey,   "key_symbols.MULTIPLY",           "npmul"),
        (vkey,   "key_symbols.ADD",                "npadd"),
        (vkey,   "key_symbols.SEPARATOR",          "npsep"),
        (vkey,   "key_symbols.SUBTRACT",           "npsub"),
        (vkey,   "key_symbols.DECIMAL",            "npdec"),
        (vkey,   "key_symbols.DIVIDE",             "npdiv"),
        (vkey,   "key_symbols.NUMPAD0",            "numpad0 np0"),
        (vkey,   "key_symbols.NUMPAD1",            "numpad1 np1"),
        (vkey,   "key_symbols.NUMPAD2",            "numpad2 np2"),
        (vkey,   "key_symbols.NUMPAD3",            "numpad3 np3"),
        (vkey,   "key_symbols.NUMPAD4",            "numpad4 np4"),
        (vkey,   "key_symbols.NUMPAD5",            "numpad5 np5"),
        (vkey,   "key_symbols.NUMPAD6",            "numpad6 np6"),
        (vkey,   "key_symbols.NUMPAD7",            "numpad7 np7"),
        (vkey,   "key_symbols.NUMPAD8",            "numpad8 np8"),
        (vkey,   "key_symbols.NUMPAD9",            "numpad9 np9")]),
    ("Function keys", 1, [
        (vkey,   "key_symbols.F1",                 "f1"),
        (vkey,   "key_symbols.F2",                 "f2"),
        (vkey,   "key_symbols.F3",                 "f3"),
        (vkey,   "key_symbols.F4",                 "f4"),
        (vkey,   "key_symbols.F5",                 "f5"),
        (vkey,   "key_symbols.F6",                 "f6"),
        (vkey,   "key_symbols.F7",                 "f7"),
        (vkey,   "key_symbols.F8",                 "f8"),
        (vkey,   "key_symbols.F9",                 "f9"),
        (vkey,   "key_symbols.F10",                "f10"),
        (vkey,   "key_symbols.F11",                "f11"),
        (vkey,   "key_symbols.F12",                "f12"),
        (vkey,   "key_symbols.F13",                "f13"),
        (vkey,   "key_symbols.F14",                "f14"),
        (vkey,   "key_symbols.F15",                "f15"),
        (vkey,   "key_symbols.F16",                "f16"),
        (vkey,   "key_symbols.F17",                "f17"),
        (vkey,   "key_symbols.F18",                "f18"),
        (vkey,   "key_symbols.F19",                "f19"),
        (vkey,   "key_symbols.F20",                "f20"),
        (vkey,   "key_symbols.F21",                "f21"),
        (vkey,   "key_symbols.F22",                "f22"),
        (vkey,   "key_symbols.F23",                "f23"),
        (vkey,   "key_symbols.F24",                "f24")]),
    ("Multimedia keys", 1, [
        (vkey,   "key_symbols.VOLUME_UP",          "volumeup volup"),
        (vkey,   "key_symbols.VOLUME_DOWN",        "volumedown voldown"),
        (vkey,   "key_symbols.VOLUME_MUTE",        "volumemute volmute"),
        (vkey,   "key_symbols.MEDIA_NEXT_TRACK",   "tracknext"),
        (vkey,   "key_symbols.MEDIA_PREV_TRACK",   "trackprev"),
        (vkey,   "key_symbols.MEDIA_PLAY_PAUSE",   "playpause"),
        (vkey,   "key_symbols.BROWSER_BACK",       "browserback"),
        (vkey,   "key_symbols.BROWSER_FORWARD",    "browserforward")]),
    )


print("# ---- Code for typeables.py")
for group_name, indent_level, group_map in key_map:
    indent = indent_level * "    "
    print("%s# %s" % (indent , group_name))
    for key_type, key_value, key_names in group_map:
        for key_name in key_names.split():
            if key_type == lookup:
                print('%s_add_typeable(name="%s", char=\'%s\')'
                      % (indent, key_name, key_value))
            elif key_type == vkey:
                value_code = ("Typeable(code=%s, name=%r)"
                              % (key_value, key_name))
                print(('%s"%s": %s%s,'
                       % (indent, key_name, " " * (16-len(key_name)),
                          value_code)))
            else:
                raise Exception("Invalid key type: {0!r} (for {1!r} {2!r})"
                                .format(key_type, key_value, key_names))
    print("")

print("# ---- Code for documentation in action_key.py")
# Exclude reserved characters in the generated documentation.
excluded = "-,:/"
for group_name, _, group_map in key_map:
    parts = [group_name + ":"]
    for key_type, key_value, key_names in group_map:
        key_name_parts = []
        for key_name in key_names.split():
            if key_name not in excluded:
                key_name = key_name.decode("string_escape")
                key_name_parts.append("``" + key_name + "``")
        parts.append(" or ".join(key_name_parts) + ",")
    parts[-1] = parts[-1][:-1]  # Remove trailing comma.
    line = " -"
    for part in parts:
        concatenation = line + " " + part
        if len(concatenation) <= 72:
            line = concatenation
        else:
            print(line)
            line = "   " + part
    print(line)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
PlaySound action
============================================================================

The :class:`PlaySound` action class is used to play wave files.


Example usage
----------------------------------------------------------------------------

The following example shows how to play a wave file using the
:class:`PlaySound` action class::

    PlaySound(file="tada.wav").execute()


Windows
----------------------------------------------------------------------------

On Windows, :class:`PlaySound` uses the `PlaySound Windows API function
<https://docs.microsoft.com/en-us/windows/win32/multimedia/the-playsound-function>`__.

The action can be used to play Windows system sounds. For example::

    # Play the system shutdown sound.
    PlaySound("SystemExit").execute()

    # Play the logout sound.
    PlaySound("WindowsLogout").execute()


System sound names are matched against registry keys.

Invalid file paths or unknown system sounds will result in the default error
sound being played. ``RuntimeErrors`` will be raised if Windows fails to
play a known system sound.


Other platforms
----------------------------------------------------------------------------

On other platforms, the :class:`PlaySound` class will use `PyAudio
<http://people.csail.mit.edu/hubert/pyaudio>`__ to play specified wave
files.

Invalid file paths will result in errors on other platforms.


Class reference
----------------------------------------------------------------------------

"""

# pylint: disable=E0401,C0413
# This file imports from optional or Win32-only packages depending on the
# platform.

from ctypes import CFUNCTYPE, c_char_p, c_int, cdll

import os
import wave

if os.name == 'nt':
    import winsound

try:
    import pyaudio
except ImportError:
    pyaudio = None

from .action_base         import ActionBase


#---------------------------------------------------------------------------

ERROR_HANDLER_FUNC = CFUNCTYPE(None, c_char_p, c_int, c_char_p, c_int,
                               c_char_p)


def _get_pa_instance():
    # Suppress initial ALSA messages if using ALSA.
    # Got this from: https://stackoverflow.com/a/17673011/12157649
    try:
        asound = cdll.LoadLibrary('libasound.so')
        c_error_handler = ERROR_HANDLER_FUNC(
            lambda filename, line, function, err, fmt: None
        )
        asound.snd_lib_error_set_handler(c_error_handler)
    except:
        # We'll most likely get here if the Port Audio host API isn't ALSA.
        asound = None

    # Create the pa instance.
    pa = pyaudio.PyAudio()

    # If necessary, restore the original error handler.
    if asound:
        asound.snd_lib_error_set_handler(None)
    return pa


def _pyaudio_play(name):
    if not name:
        return

    if pyaudio is None:
        # Raise an error because pyaudio isn't installed.
        raise RuntimeError("pyaudio must be installed to use PlaySound "
                           "on this platform")

    # Play the wave file.
    pa = _get_pa_instance()
    chunk = 1024
    wf = wave.open(name, 'rb')
    stream = pa.open(format=pa.get_format_from_width(wf.getsampwidth()),
                     channels=wf.getnchannels(), rate=wf.getframerate(),
                     output=True)
    data = wf.readframes(chunk)
    while data:
        stream.write(data)
        data = wf.readframes(chunk)

    stream.stop_stream()
    stream.close()
    pa.terminate()


class PlaySound(ActionBase):
    """
        Start playing a wave file or system sound.

        When this action is executed, the specified wave file or named
        system sound is played.

        Playing named system sounds is only supported on Windows.

    """

    def __init__(self, name='', file=None):
        """
            Constructor arguments:
             - *name* (*str*, default *empty string*) --
               name of the Windows system sound to play. This argument is
               effectively an alias for *file* on other platforms.
             - *file* (*str*, default *None*) --
               path of wave file to play when the action is executed.

            If *name* and *file* are both *None*, then waveform playback
            will be silenced on Windows when the action is executed. Nothing
            will happen on other platforms.

        """
        ActionBase.__init__(self)
        self._flags = 0
        if file is not None:
            self._name = file
            if os.name == 'nt':
                self._flags = winsound.SND_FILENAME
        else:
            self._name = name
            if os.name == 'nt':
                self._flags = winsound.SND_ALIAS

        # Expand ~ constructions and shell variables in the file path if
        # necessary.
        if file is not None or os.name != 'nt':
            self._name = os.path.expanduser(os.path.expandvars(self._name))

        self._str = str(self._name)

    def _execute(self, data=None):
        if os.name == 'nt':
            # Play the file or sound using the Windows API.
            winsound.PlaySound(self._name, self._flags)
        else:
            # Play the file with pyaudio.
            _pyaudio_play(self._name)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# pylint: disable=E0401
# This file imports Win32-only symbols.

from ctypes import windll, pointer, c_long, c_ulong, Structure

import win32con

from dragonfly.actions.sendinput import (MouseInput, make_input_array,
                                         send_input_array)
from ._base import BaseButtonEvent, MoveEvent


#---------------------------------------------------------------------------
# Functions and event delegate for getting and setting the cursor position.

class Point(Structure):
    _fields_ = [
        ('x',  c_long),
        ('y',  c_long),
    ]


def get_cursor_position():
    point = Point()
    result = windll.user32.GetCursorPos(pointer(point))
    if result:
        return point.x, point.y
    else:
        return None


def set_cursor_position(x, y):
    result = windll.user32.SetCursorPos(c_long(int(x)), c_long(int(y)))
    return not result


class MoveEventDelegate(object):

    @classmethod
    def get_position(cls):
        return get_cursor_position()

    @classmethod
    def set_position(cls, x, y):
        return set_cursor_position(x, y)


# Set MoveEvent's delegate. This allows us to set platform-specific
# functions for getting and setting the cursor position without having
# to do a lot of sub-classing for no good reason.
MoveEvent.delegate = MoveEventDelegate()

#---------------------------------------------------------------------------
# Win32 mouse button and wheel up/down flags.

# Taken from https://msdn.microsoft.com/en-us/library/windows/desktop/ms646273(v=vs.85).aspx
MOUSEEVENTF_HWHEEL = 0x1000

PLATFORM_BUTTON_FLAGS = {
    "left":   ((win32con.MOUSEEVENTF_LEFTDOWN, 0),
               (win32con.MOUSEEVENTF_LEFTUP, 0)),
    "right":  ((win32con.MOUSEEVENTF_RIGHTDOWN, 0),
               (win32con.MOUSEEVENTF_RIGHTUP, 0)),
    "middle": ((win32con.MOUSEEVENTF_MIDDLEDOWN, 0),
               (win32con.MOUSEEVENTF_MIDDLEUP, 0)),
    "four": ((win32con.MOUSEEVENTF_XDOWN, 1),
             (win32con.MOUSEEVENTF_XUP, 1)),
    "five": ((win32con.MOUSEEVENTF_XDOWN, 2),
             (win32con.MOUSEEVENTF_XUP, 2)),
}

PLATFORM_WHEEL_FLAGS = {
    "wheelup": (win32con.MOUSEEVENTF_WHEEL, 120),
    "stepup": (win32con.MOUSEEVENTF_WHEEL, 40),
    "wheeldown": (win32con.MOUSEEVENTF_WHEEL, -120),
    "stepdown": (win32con.MOUSEEVENTF_WHEEL, -40),
    "wheelright": (MOUSEEVENTF_HWHEEL, 120),
    "stepright": (MOUSEEVENTF_HWHEEL, 40),
    "wheelleft": (MOUSEEVENTF_HWHEEL, -120),
    "stepleft": (MOUSEEVENTF_HWHEEL, -40),
}


#---------------------------------------------------------------------------
# Win32 event classes.

class ButtonEvent(BaseButtonEvent):

    def execute(self, window):
        # Ensure that the primary mouse button is the *left* button before
        #  sending events.
        primary_changed = windll.user32.SwapMouseButton(0)
        try:
            # Prepare and send the mouse events.
            zero = pointer(c_ulong(0))
            inputs = [MouseInput(0, 0, flag[1], flag[0], 0, zero)
                      for flag in self._flags]
            array = make_input_array(inputs)
            send_input_array(array)
        finally:
            # Swap the primary mouse button back if it was previously set
            #  to *right*.
            if primary_changed:
                windll.user32.SwapMouseButton(1)



# -*- encoding: windows-1252 -*-
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import unittest

from six import PY2

from dragonfly.actions.action_base import Repeat
from dragonfly.actions.action_function import Function
from dragonfly.actions.action_key import Key
from dragonfly.actions.action_mimic import Mimic
from dragonfly.actions.action_paste import Paste
from dragonfly.actions.action_text import Text


#===========================================================================

class TestNonAsciiText(unittest.TestCase):

    def test_non_ascii_text(self):
        """ Test handling of non-ASCII characters in Text action. """

        action = Text(u"touché")
        self.assertEqual(str(action), "%r" % (u"touché",))


class TestNonAsciiKey(unittest.TestCase):

    def test_non_ascii_key(self):
        """ Test handling of non-ASCII characters in Key action. """

        action = Key(u"é")
        self.assertEqual(str(action), "[%r]" % (u"é",))


class TestNonAsciiActionSeries(unittest.TestCase):

    def test_non_ascii_key(self):
        """ Test handling of non-ASCII characters in ActionSeries actions.
        """

        action = Key(u"é") + Text(u"á")
        self.assertEqual(str(action), "[%r]+%r" % (u"é", u"á"))

        action = Key(u"é") | Text(u"á")
        self.assertEqual(str(action), "[%r]|%r" % (u"é", u"á"))


class TestNonAsciiFunction(unittest.TestCase):

    def test_non_ascii_function(self):
        """ Test handling of non-ASCII characters in Function action. """

        action = Function(lambda: u"é")
        expected = u"\\xe9" if PY2 else u"é"
        self.assertIn(expected, str(action))


class TestNonAsciiPaste(unittest.TestCase):

    def test_non_ascii_paste(self):
        """ Test handling of non-ASCII characters in Paste action. """

        action = Paste("touché")
        self.assertEqual(str(action), "Paste(%r)" % ("touché",))


class TestNonAsciiMimic(unittest.TestCase):

    def test_non_ascii_mimic(self):
        """ Test handling of non-ASCII characters in Mimic action. """

        action = Mimic("touché")
        self.assertEqual(str(action), "Mimic(%r)" % ("touché",))


class TestRepeat(unittest.TestCase):

    def test_repeat(self):
        """ Test handling of Repeat elements """

        r1 = Repeat("n")
        self.assertEqual(r1.factor({"n": 3}), 3)
        r2 = Repeat("n", 3)
        self.assertEqual(r2.factor({"n": 3}), 6)
        r3 = Repeat(3)
        self.assertEqual(r3.factor(), 3)
        r4 = Repeat(3, "n")
        self.assertEqual(r4.factor({"n": 3}), 6)


#===========================================================================

if __name__ == "__main__":
    unittest.main()



from six import text_type

from dragonfly import (
    Key,
    Mouse,
    Text,
)

from . import base
from . import utils


class AccessibilityController(object):
    """OS-independent controller for accessing accessibility functionality."""

    def __init__(self, os_controller):
        self.os_controller = os_controller
        self.os_controller.start()
        self.stopped = False

    def stop(self):
        """Stops the controller (otherwise process exit may be blocked)."""

        self.os_controller.stop()
        self.stopped = True

    def is_editable_focused(self):
        """True if an editable text field is focused."""

        return utils.is_editable_focused(self.os_controller)

    def move_cursor(self, text_query, position):
        """Moves the cursor before or after text that matches the provided query."""

        return utils.move_cursor(self.os_controller, text_query, position)

    def select_text(self, text_query):
        """Selects text which matches the provided query."""

        try:
            return utils.select_text(self.os_controller, text_query)
        except base.UnsupportedSelectionError:
            # Fall back to mouse-based text selection.
            text_info = utils.get_text_info(self.os_controller, text_query)
            if text_info and text_info.start_coordinates and text_info.end_coordinates:
                Mouse("[%d, %d], left:down, [%d, %d]/10, left:up" %
                      (text_info.start_coordinates + text_info.end_coordinates)).execute()
                return True
            else:
                return False

    def replace_text(self, text_query, replacement):
        """Replaces text which matches the provided query."""

        saved_cursor = utils.get_cursor_offset(self.os_controller)
        text_info = utils.get_text_info(self.os_controller, text_query)
        if not text_info:
            return
        cursor_before = text_info.end
        if self.select_text(text_query):
            # Replace text.
            if replacement:
                replacement = text_type(replacement).lower()
                if text_info.text and text_info.text.isupper():
                    replacement = replacement.upper()
                elif text_info.text and text_info.text[0].isupper():
                    replacement = replacement.capitalize()
                Text(replacement.replace("%", "%%")).execute()
            else:
                # Delete the selected text with a single backspace. Note that
                # the user may want to delete the preceding whitespace as well,
                # and in order to do that they should simply include that
                # ("space bar") in the text they dictate. The problem is that
                # some editors (e.g. Google Docs) do this automatically, so we
                # can't simply always use 2 backspaces when a word is deleted.
                Key("backspace").execute()

            # Restore cursor position.
            if saved_cursor is not None:
                if saved_cursor <= text_info.start:
                    utils.set_cursor_offset(self.os_controller, saved_cursor)
                elif saved_cursor <= text_info.end:
                    # Saved cursor is within changed region, leave as-is.
                    pass
                else:
                    cursor_after = utils.get_cursor_offset(self.os_controller)
                    if cursor_after:
                        utils.set_cursor_offset(self.os_controller, saved_cursor + cursor_after - cursor_before)



# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
RPC methods
----------------------------------------------------------------------------

For RPC methods to work they must be added to the server with
:meth:`add_method`. For example::

    from dragonfly.engines import get_engine
    from dragonfly.rpc import RPCServer

    # Initialise and start the server.
    server = RPCServer()
    server.start()

    # Add the RPC method via decoration.
    @server.add_method
    def get_engine_language():
        return get_engine().language

    # add_method() can also be used normally.
    server.add_method(get_engine_language)


Sending requests
============================================================================
Requests can be sent to the server using, the :meth:`send_rpc_request`
function from Python::

    send_rpc_request(
        server.url, method="get_engine_language",
        params=[server.security_token], id=0
    )

Other tools such as `curl <https://curl.haxx.se/>`_ can also be used.

..  code:: shell

    Using positional arguments:
    $ curl --data-binary '{"jsonrpc":"2.0","id": "0","method": "speak","params": ["hello world", "<security-token>"]}' -H 'content-type:text/json;' http://127.0.0.1:50051

    Using key word arguments:
    $ curl --data-binary '{"jsonrpc":"2.0","id": "0","method": "speak","params": {"text": "hello world", "security_token": "<security-token>"}}' -H 'content-type:text/json;' http://127.0.0.1:50051


Built-in RPC methods
============================================================================

"""

from six import string_types

from dragonfly.grammar.recobs import RecognitionHistory
from .. import get_engine, CompoundRule, MappingRule
from ..engines.base import MimicFailure


methods = []


def _add_method(m):
    methods.append(m)
    return m


@_add_method
def list_grammars():
    """
    Get a list of grammars loaded into the current engine.

    This includes grammar rules and attributes.
    """
    # Send rule and grammar data back to the client.
    grammars = []
    engine = get_engine()
    for grammar in engine.grammars:
        rules = []
        for rule in grammar.rules:
            if isinstance(rule, CompoundRule):
                specs = [rule.spec]
            elif isinstance(rule, MappingRule):
                specs = sorted(rule.specs)
            else:
                specs = [rule.element.gstring()]
            rules.append({
                "name": rule.name, "specs": specs,
                "exported": rule.exported, "active": rule.active
            })

        # Add grammar attributes to the grammars list.
        is_active = any([r.active for r in grammar.rules])
        grammars.append({"name": grammar.name, "rules": rules,
                         "enabled": grammar.enabled, "active": is_active})

    return grammars


@_add_method
def mimic(words):
    """
    Mimic the given *words*.

    :param words: string or list of words to mimic
    :returns: whether the mimic was a success
    :rtype: bool
    """
    if isinstance(words, string_types):
        words = words.split()
    try:
        get_engine().mimic(words)
        return True
    except MimicFailure:
        return False


@_add_method
def speak(text):
    """
    Speak the given *text* using text-to-speech using :meth:`engine.speak`.

    :param text: text to speak using text-to-speech
    :type text: str
    """
    if not isinstance(text, string_types):
        raise TypeError("text must be a string")

    return get_engine().speak(text)


@_add_method
def get_engine_language():
    """
    Get the current engine's language.

    :returns: language code
    :rtype: str
    """
    return get_engine().language


_history_closure = [None]

class RPCRecognitionHistory(RecognitionHistory):
    """"""

    def __init__(self, length=10, record_failures=False):
        RecognitionHistory.__init__(self, length)
        self._record_failures = record_failures

    def on_failure(self):
        # Include recognition failures if requested.
        if self._record_failures:
            self.append(False)
            if self._length:
                while len(self) > self._length:
                    self.pop(0)


@_add_method
def register_history(length=10, record_failures=False):
    """
    Register an internal recognition observer.

    :param length: length to initialize the ``RecognitionHistory`` instance
        with (default ``10``).
    :param record_failures: whether to record recognition failures
        (default ``False``).
    :type record_failures: bool
    :type length: int
    """
    if _history_closure[0] is not None:
        raise RuntimeError("history has already been registered")

    obs = RPCRecognitionHistory(length, record_failures)
    obs.register()
    _history_closure[0] = obs


@_add_method
def get_recognition_history():
    """
    Get the recognition history if an observer is registered.

    The :meth:`register_history` method **must** be called to register the
    observer first.

    :returns: history
    :rtype: list
    """
    obs = _history_closure[0]
    if obs is None:
        raise RuntimeError("the recognition observer is not registered")

    return list(obs)


@_add_method
def is_in_speech():
    """
    Whether the user is currently speaking.

    The :meth:`register_history` method **must** be called to register the
    observer first.

    :rtype: bool
    """
    obs = _history_closure[0]
    if obs is None:
        raise RuntimeError("the recognition observer is not registered")

    return not obs.complete


@_add_method
def unregister_history():
    """
    Unregister the internal recognition observer.
    """
    if _history_closure[0] is None:
        raise RuntimeError("the recognition observer is not registered")

    _history_closure[0].unregister()
    _history_closure[0] = None



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import logging

from six import integer_types

from .rectangle import Rectangle


#===========================================================================
# Monitor classes for storing info about a single display monitor.

class BaseMonitor(object):
    """
    The base monitor class.
    """

    _log = logging.getLogger("monitor.init")

    #-----------------------------------------------------------------------
    # Class attributes to retrieve existing Monitor objects.

    _monitors_by_handle = {}

    #-----------------------------------------------------------------------
    # Class methods to create new Monitor objects.

    @classmethod
    def get_monitor(cls, handle, rectangle):
        """
        Get a :class:`Monitor` object given a monitor handle.

        Given the same handle, this method will return the same object.

        :param handle: monitor handle
        :type handle: int
        :param rectangle: monitor rectangle
        :type rectangle: Rectangle
        :rtype: Monitor
        :returns: monitor
        """
        if handle in cls._monitors_by_handle:
            monitor = cls._monitors_by_handle[handle]
            monitor.rectangle = rectangle
        else:
            monitor = cls(handle, rectangle)
            cls._monitors_by_handle[handle] = monitor
        return monitor

    @classmethod
    def get_all_monitors(cls):
        """
        Get a list containing each connected monitor.

        :rtype: list
        :returns: monitors
        """
        raise NotImplementedError()

    #-----------------------------------------------------------------------
    # Methods for initialization and introspection.

    def __init__(self, handle, rectangle):
        assert isinstance(handle, integer_types)
        self._handle = handle
        assert isinstance(rectangle, Rectangle)
        self._rectangle = rectangle

    def __repr__(self):
        return "%s(%d)" % (self.__class__.__name__, self._handle)

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    def _set_handle(self, handle):
        assert isinstance(handle, integer_types)
        self._handle = handle

    handle = property(fget=lambda self: self._handle,
                      fset=_set_handle,
                      doc="Protected access to handle attribute.")

    def _set_rectangle(self, rectangle):
        assert isinstance(rectangle, Rectangle)
        self._rectangle = rectangle  # Should make a copy??

    rectangle = property(fget=lambda self: self._rectangle,
                         fset=_set_rectangle,
                         doc="Protected access to rectangle attribute.")

    @property
    def is_primary(self):
        """
        Whether this is the primary display monitor.

        :rtype: bool
        :returns: true or false
        """
        rect = self.rectangle
        return rect.x == 0 and rect.y == 0

    @property
    def name(self):
        """
        The name of this monitor.

        :rtype: str
        :returns: monitor name
        """
        raise NotImplementedError()


class FakeMonitor(BaseMonitor):
    """
    The monitor class used on unsupported platforms.
    """

    @classmethod
    def get_all_monitors(cls):
        return []

    @property
    def name(self):
        """
        The name of this monitor.

        :rtype: str
        :returns: monitor name
        """
        return self.__class__.__name__



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements a Point class for geometry operations.
"""

# pylint: disable=W0212
# Suppress warnings about protected member access (e.g for '_x').

import math

from six import integer_types

#===========================================================================

class Point(object):

    #-----------------------------------------------------------------------
    # Methods for initialization, copying, and introspection.

    def __init__(self, x=None, y=None):
        self._x = 0.0; self._y = 0.0
        if x is not None: self.x = x
        if y is not None: self.y = y

    def copy(self):     return Point(x=self._x, y=self._y)
    def __copy__(self): return self.copy()

    def __repr__(self):
        return "%s(%.2f,%.2f)" % (self.__class__.__name__, self._x, self._y)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return not self.__eq__(other)

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    def _set_x(self, x):
        if isinstance(x, float):    self._x = x
        elif isinstance(x, integer_types):    self._x = float(x)
        else: raise TypeError("Point coordinate must be an int or float;"
                              " received %r." % x)
    x = property(fget=lambda self: self._x,
                 fset=_set_x,
                 doc="Protected access to x attribute.")

    def _set_y(self, y):
        if isinstance(y, float):    self._y = y
        elif isinstance(y, integer_types):    self._y = float(y)
        else: raise TypeError("Point coordinate must be an int or float;"
                              " received %r." % y)
    y = property(fget=lambda self: self._y,
                 fset=_set_y,
                 doc="Protected access to y attribute.")

    @property
    def magnitude(self):
        return math.hypot(self._x, self._y)

    #-----------------------------------------------------------------------
    # Methods for manipulating point objects.

    def __iadd__(self, other):
        """Translate point by adding another point's coordinates."""
        if not isinstance(other, Point): return NotImplemented
        self._x += other._x
        self._y += other._y
        return self

    def __add__(self, other):
        """Create a new point with coordinates of this point and other
            added up."""
        if not isinstance(other, Point): return NotImplemented
        clone = self.copy()
        clone += other
        return clone

    def __isub__(self, other):
        """Translate point by adding another point's coordinates."""
        if not isinstance(other, Point): return NotImplemented
        self._x -= other._x
        self._y -= other._y
        return self

    def __sub__(self, other):
        """Create a new point with coordinates of this point and other
            added up."""
        if not isinstance(other, Point): return NotImplemented
        clone = self.copy()
        clone -= other
        return clone

    def translate(self, dx, dy):
        """camp point."""
        # pylint: disable=R0201
        # Suppress no-self-use warning.
        other = Point(dx, dy)
        self += other

    def interpolate(self, other, fraction):
        """Create a new point with coordinates interpolated between
            this point and another."""
        if not isinstance(other, Point):
            raise TypeError("Can only interpolate between points;"
                            " received %r" % other)
        x1 = self._x; y1 = self._y; x2 = other._x; y2 = other._y
        x3 = x1 + (x2 - x1) * fraction
        y3 = y1 + (y2 - y1) * fraction
        return Point(x3, y3)

    def average(self, other):
        """Translate point."""
        return self.interpolate(other, 0.5)

    def renormalize(self, src, dst):
#       assert isinstance(src, Rectangle)
#       assert isinstance(dst, Rectangle)
        nx = (self._x - src.x1) / src.dx
        self._x = dst.x1 + nx * dst.dx
        ny = (self._y - src.y1) / src.dy
        self._y = dst.y1 + ny * dst.dy
        return self



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Multiplexing interface to a timer
============================================================================

"""

import time
import logging

from threading import Thread, current_thread, RLock

#---------------------------------------------------------------------------

class Timer(object):
    """
    Timer class for calling a function every N seconds.

    Constructor arguments:

     - *function* (*callable*) -- the function to call every N seconds. Must
       have no required arguments.
     - *interval* (*float*) -- number of seconds between calls to the
       function. Note that this is on a best-effort basis only.
     - *manager* (:class:`TimerManagerBase`) -- engine timer manager instance.
     - *repeating* (*bool*) -- whether to call the function every N seconds
       or just once (default: True).

    Instances of this class are normally initialised from
    :meth:`engine.create_timer`.
    """

    _log = logging.getLogger("engine.timer")

    def __init__(self, function, interval, manager, repeating=True):
        self.function = function
        self.interval = interval
        self.manager = manager
        self.repeating = repeating
        self.active = False
        self.next_time = None
        self.start()

    def start(self):
        """
        Start calling the timer's function on an interval.

        This method is called on initialisation.
        """
        if self.active:
            return
        self.manager.add_timer(self)
        self.active = True
        self.next_time = time.time() + self.interval

    def stop(self):
        """ Stop calling the timer's function on an interval. """
        if not self.active:
            return
        self.manager.remove_timer(self)
        self.active = False
        self.next_time = None

    def call(self):
        """
        Call the timer's function.

        This method is normally called by the timer manager.
        """
        self.next_time = time.time() + self.interval
        try:
            self.function()
        except Exception as e:
            self._log.exception("Exception during timer callback: %s" % (e,))
        if not self.repeating:
            self.stop()


class TimerManagerBase(object):
    """ Base timer manager class. """

    _log = logging.getLogger("engine.timer")

    def __init__(self, interval, engine):
        self.interval = interval
        self.engine = engine
        self.timers = []
        self._lock = RLock()
        self._enabled = True
        self._active = False

    def add_timer(self, timer):
        """ Add a timer and activate the main callback if required. """
        with self._lock:
            self.timers.append(timer)
        if len(self.timers) == 1 and self._enabled:
            self._activate_main_callback(self.main_callback,
                                         self.interval)
            self._active = True

    def remove_timer(self, timer):
        """ Remove a timer and deactivate the main callback if required. """
        try:
            with self._lock:
                self.timers.remove(timer)
        except Exception as e:
            self._log.exception("Failed to remove timer: %s" % e)
            return
        if len(self.timers) == 0 and self._enabled:
            self._deactivate_main_callback()
            self._active = False

    def enable(self):
        """
        Method to re-enable the main timer callback.

        The main timer callback is enabled by default. This method is only
        useful if :meth:`disable` is called.
        """
        self._enabled = True

    def disable(self):
        """
        Method to disable execution of the main timer callback.

        This method is used for testing timer-related functionality without
        race conditions.
        """
        self._enabled = False

    def main_callback(self):
        """ Method to call each timer's function when required. """
        # Deactivate the engine's main callback if necessary. We don't
        # return early here because this method should work if called
        # manually.
        if not self._enabled and self._active:
            self._deactivate_main_callback()
            self._active = False

        now = time.time()
        for c in tuple(self.timers):
            if c.next_time < now:
                try:
                    c.call()
                except Exception as e:
                    self._log.exception("Exception occurred during"
                                        " timer callback: %s" % (e,))

    def _activate_main_callback(self, callback, msec):
        """
        Virtual method to implement to start calling :meth:`main_callback`
        on an interval.
        """
        raise NotImplementedError("_activate_main_callback() not implemented for"
                                  " %s class." % self.__class__.__name__)

    def _deactivate_main_callback(self):
        """
        Virtual method to implement to stop calling :meth:`main_callback` on
        an interval.
        """
        raise NotImplementedError("_deactivate_main_callback() not implemented for"
                                  " %s class." % self.__class__.__name__)


class ThreadedTimerManager(TimerManagerBase):
    """
    Timer manager class using a daemon thread.

    This class is used by the "text" engine. It is only suitable for engine
    backends with no recognition loop to execute timer functions on.

    .. warning::

       The timer interface is **not** thread-safe. Use the :meth:`enable`
       and :meth:`disable` methods if you need finer control over timer
       function execution.
    """

    def __init__(self, interval, engine):
        TimerManagerBase.__init__(self, interval, engine)
        self._running = False
        self._thread = None

    def _activate_main_callback(self, callback, sec):
        """"""
        # Do nothing if the thread is already running.
        if self._running:
            return

        def run():
            while self._running:
                time.sleep(sec)
                callback()

        self._running = True
        self._thread = Thread(target=run)
        self._thread.setDaemon(True)
        self._thread.start()

    def _deactivate_main_callback(self):
        """"""
        # Stop the thread's main loop and wait until it finishes, timing out
        # after 5 seconds.
        should_join = (self._thread and self._thread.is_alive() and
                       self._thread is not current_thread())
        self._running = False
        if should_join:
            timeout = 5
            self._thread.join(timeout=timeout)
            if self._thread.is_alive():
                raise RuntimeError("failed to deactivate main callback "
                                   "after %d seconds" % timeout)


class DelegateTimerManagerInterface(object):
    """
    DelegateTimerManager interface.
    """

    def __init__(self):
        self._timer_callback = None
        self._timer_interval = None
        self._timer_next_time = 0

    def set_timer_callback(self, callback, sec):
        """
        Method to set the timer manager's callback.

        :param callback: function to call every N seconds
        :type callback: callable | None
        :param sec: number of seconds between calls to the callback function
        :type sec: float | int
        """
        self._timer_callback = callback
        self._timer_interval = sec
        self._timer_next_time = time.time()

    def call_timer_callback(self):
        """"""
        if not (self._timer_callback and self._timer_interval):
            return

        now = time.time()
        if self._timer_next_time < now:
            self._timer_next_time = now + self._timer_interval
            self._timer_callback()


class DelegateTimerManager(TimerManagerBase):
    """
    Timer manager class that calls :meth:`main_callback` through an
    engine-specific callback function.

    Engines using this class should implement the methods in
    :class:`DelegateManagerInterface`.

    This class is used by the SAPI 5 engine.
    """

    def __init__(self, interval, engine):
        TimerManagerBase.__init__(self, interval, engine)

    def _activate_main_callback(self, callback, sec):
        """"""
        self.engine.set_timer_callback(callback, sec)

    def _deactivate_main_callback(self):
        """"""
        self.engine.set_timer_callback(None, 0)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import locale
from subprocess import Popen, PIPE

from six import string_types, binary_type

from .base_monitor import BaseMonitor
from .rectangle import Rectangle


#===========================================================================
# Monitor class for storing info about a single display monitor.

class X11Monitor(BaseMonitor):
    """
    The monitor class used on X11 (Linux).

    This implementation parses monitor information from
    ``xrandr``.
    """

    #-----------------------------------------------------------------------
    # Methods for initialization and introspection.

    def __init__(self, name, rectangle):
        assert isinstance(name, string_types)
        self._name = name
        self._primary = False

        # Get a numeric value from the name that we can use as a handle.
        handle = hash(name)
        super(X11Monitor, self).__init__(handle, rectangle)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self._name)

    #-----------------------------------------------------------------------
    # Class methods to create new Monitor objects.

    @classmethod
    def get_all_monitors(cls):
        # pylint: disable=R0912,R0914,R0915
        # Suppress warnings about too many local variables.

        # Get updated monitor information from xrandr.
        try:
            p = Popen(["xrandr"], stdout=PIPE)
            stdout, _ = p.communicate()

            # Decode output if it is binary.
            encoding = locale.getpreferredencoding()
            if isinstance(stdout, binary_type):
                stdout = stdout.decode(encoding)

            # Handle non-zero return codes.
            if p.wait() > 0:
                print(stdout)
                raise RuntimeError("xrandr exited with non-zero return "
                                   "code %d" % p.returncode)
        except Exception as e:
            cls._log.error("Failed to execute command: %s. Is "
                           "xrandr installed?", e)
            raise e

        # Parse output from xrandr.
        monitors = []
        for line in stdout.split("\n")[1:]:  # skip the first line
            # Skip disconnected monitors, output modes and empty lines.
            if not line or "disconnected" in line or line.startswith("   "):
                continue

            # Get the information we need.
            parts = line.split()
            name = parts.pop(0)
            primary = "primary" in parts
            if primary:  # this monitor will be first in the list
                parts.remove("primary")

            # Find and remove info we don't need (e.g. 'connected' and
            # orientation info).
            i = 1
            for i, part in enumerate(parts):
                if "(" in part:  # e.g. '(normal'
                    break
            parts = parts[1:i]

            # Skip disabled monitors (no reported mode info).
            if not parts:
                continue

            # Get resolution and position info.
            res_info = parts[0].split("+")
            dx, dy = res_info.pop(0).split("x")
            top_left_x, top_left_y = res_info

            # Convert values to integers and create a rectangle object
            # representing the monitor's dimensions and relative position.
            dx, dy = int(dx), int(dy)
            top_left_x, top_left_y = int(top_left_x), int(top_left_y)
            rectangle = Rectangle(top_left_x, top_left_y, dx, dy)

            # Get a new or updated monitor object and add it to the list.
            monitor = cls.get_monitor(name, rectangle)
            monitor.is_primary = primary

            # Ensure that the top_left monitor is the first in the list.
            if top_left_x == 0:
                monitors.insert(0, monitor)
            else:
                monitors.append(monitor)

        # Return the list of monitors.
        return monitors

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    def _set_primary(self, value):
        self._primary = bool(value)

    is_primary = property(fget=lambda self: self._primary,
                          fset=_set_primary, doc="Whether this is the "
                          "primary display monitor.")

    def _set_name(self, name):
        assert isinstance(name, string_types)
        self._name = name

        # Also set the handle, as it is derived from the name.
        self._set_handle(hash(name))

    name = property(fget=lambda self: self._name,
                    fset=_set_name,
                    doc="Protected access to name attribute.")



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Dutch language implementations of Integer and Digits classes
============================================================================

"""

from ..base.integer_internal  import (MapIntBuilder, CollectionIntBuilder,
                                      MagnitudeIntBuilder, IntegerContentBase)
from ..base.digits_internal   import DigitsContentBase


#---------------------------------------------------------------------------

int_0           = MapIntBuilder({
                                 "nul": 0
                               })
int_1_9         = MapIntBuilder({
                                 "1":      1,  # Use "1" instead of "een"
                                               #  because the latter has
                                               #  multiple pronunciations.
                                 "twee":   2,
                                 "drie":   3,
                                 "vier":   4,
                                 "vijf":   5,
                                 "zes":    6,
                                 "zeven":  7,
                                 "acht":   8,
                                 "negen":  9,
                               })
int_10_19       = MapIntBuilder({
                                 "tien":      10,
                                 "elf":       11,
                                 "twaalf":    12,
                                 "dertien":   13,
                                 "veertien":  14,
                                 "vijftien":  15,
                                 "zestien":   16,
                                 "zeventien": 17,
                                 "achtien":   18,
                                 "negentien": 19,
                               })
int_20_90_10    = MapIntBuilder({
                                 "twintig":   2,
                                 "dertig":    3,
                                 "veertig":   4,
                                 "vijftig":   5,
                                 "zestig":    6,
                                 "zeventig":  7,
                                 "tachtig":   8,
                                 "negentig":  9,
                               })
int_20_99       = MagnitudeIntBuilder(
                   factor      = 10,
                   spec        = "[<remainder> en] <multiplier>",
                   multipliers = [int_20_90_10],
                   remainders  = [int_1_9]
                  )
int_en_1_99     = CollectionIntBuilder(
                   spec        = "[en] <element>",
                   set         = [int_1_9, int_10_19, int_20_99]
                  )
int_100s        = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] honderd [<remainder>]",
                   multipliers = [int_1_9],
                   remainders  = [int_en_1_99]
                  )
int_100big      = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] honderd [<remainder>]",
                   multipliers = [int_10_19, int_20_99],
                   remainders  = [int_en_1_99]
                  )
int_1000s       = MagnitudeIntBuilder(
                   factor      = 1000,
                   spec        = "[<multiplier>] duizend [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99, int_100s],
                   remainders  = [int_en_1_99, int_100s]
                  )
int_1000000s    = MagnitudeIntBuilder(
                   factor      = 1000000,
                   spec        = "[<multiplier>] millioen [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99, int_100s, int_1000s],
                   remainders  = [int_en_1_99, int_100s, int_1000s]
                  )


#---------------------------------------------------------------------------

class IntegerContent(IntegerContentBase):
    builders = [int_0, int_1_9, int_10_19, int_20_99,
                int_100s, int_100big, int_1000s, int_1000000s]

class DigitsContent(DigitsContentBase):
    digits = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]



"""
NatLink importable rules dgndictation, dgnletters and dgnwords
===============================================================================

This file is a showcase of the three importable rules provided by 
NatLink: "dgndictation", "dgnletters" and "dgnwords". 

The test code below makes use of Dragonfly's built-in element testing 
tool. When run, it will connect to the speech recognition engine, load 
the element being tested, mimic recognitions, and process the recognized 
value. 

"""


#---------------------------------------------------------------------------

import doctest
from dragonfly                          import *
from dragonfly.test.infrastructure      import RecognitionFailure
from dragonfly.test.element_testcase    import ElementTestCase
from dragonfly.test.element_tester      import ElementTester


#---------------------------------------------------------------------------

class RecognitionAnalysisRule(CompoundRule):
    """
        Base class that implements reporting in human-readable format
        details about the recognized phrase. It is used by the actual
        testing rules below, and allows the doctests above to be easily
        readable and informative.

    """

    def _process_recognition(self, node, extras):
        Paste(text).execute()

    def value(self, node):
        return self.get_recognition_info(node)

    def get_recognition_info(self, node):
        output = []
        output.append('Recognition: "{0}"'.format(" ".join(node.words())))
        output.append("Word and rule pairs:")
        for word, rule in node.full_results():
            output.append(" - {0} ({1})".format(word, rule))
        output.append("Extras:")
        for key in sorted(extra.name for extra in self.extras):
            extra_node = node.get_child_by_name(key)
            if extra_node:
                output.append(" - {0}: {1}".format(key, extra_node.value()))
        return "\n".join(output)


#---------------------------------------------------------------------------

class DgnImported(RuleRef):
    """
        Base class that implements an imported rule and takes care of
        decoding recognitions accordingly.

    """

    def __init__(self, name=None, imported_name=None, default=None):
        if not imported_name:
            imported_name = name
        self.imported_name = imported_name
        self.imported_rule = Rule(self.imported_name, imported=True)
        RuleRef.__init__(self, self.imported_rule, name, default=default)

    def decode(self, state):
        state.decode_attempt(self)

        # Check that at least one word has been dictated, otherwise fail.
        if state.rule() != self.imported_name:
            state.decode_failure(self)
            return

        # Determine how many words have been dictated.
        count = 1
        while state.rule(count) == self.imported_name:
            count += 1

        # Yield possible states where the number of dictated words
        # gobbled is decreased by 1 between yields.
        for i in range(count, 0, -1):
            state.next(i)
            state.decode_success(self)
            yield state
            state.decode_retry(self)
            state.decode_rollback(self)

        # None of the possible states were accepted, failure.
        state.decode_failure(self)
        return

    def value(self, node):
        return node.words()


#---------------------------------------------------------------------------

class DgnDictationRule(RecognitionAnalysisRule):
    """
        >>> tester = ElementTester(RuleRef(DgnDictationRule()))
        >>> print(tester.recognize("dictation hello world"))
        Recognition: "dictation hello world"
        Word and rule pairs:
         - dictation (1)
         - hello (1000000)
         - world (1000000)
        Extras:
         - dgndictation: [u'hello', u'world']

        >>> print(tester.recognize("dictation"))
        RecognitionFailure

    """

    spec   = "dictation <dgndictation>"
    extras = [DgnImported("dgndictation")]


#---------------------------------------------------------------------------

class DgnLettersRule(RecognitionAnalysisRule):
    """
        >>> tester = ElementTester(RuleRef(DgnLettersRule()))
        >>> print(tester.recognize("letters a b c"))
        Recognition: "letters a\\\\l b\\\\l c\\letter-c\\l"
        Word and rule pairs:
         - letters (1)
         - a\\\\l (1000001)
         - b\\\\l (1000001)
         - c\\letter-c\\l (1000001)
        Extras:
         - dgnletters: [u'a\\\\\\\\l', u'b\\\\\\\\l', u'c\\\\letter-c\\\\l']

        >>> print(tester.recognize("letters hello world"))
        RecognitionFailure

        >>> print(tester.recognize("letters"))
        RecognitionFailure

    """

    spec   = "letters <dgnletters>"
    extras = [DgnImported("dgnletters")]


#---------------------------------------------------------------------------

class DgnWordsRule(RecognitionAnalysisRule):
    """
        >>> tester = ElementTester(RuleRef(DgnWordsRule()))
        >>> print(tester.recognize("words hello"))
        Recognition: "words hello"
        Word and rule pairs:
         - words (1)
         - hello (3)
        Extras:
         - dgnwords: [u'hello']

        >>> print(tester.recognize("words hello world"))
        RecognitionFailure

        >>> print(tester.recognize("words"))
        RecognitionFailure

    """

    spec   = "words <dgnwords>"
    extras = [DgnImported("dgnwords")]


#---------------------------------------------------------------------------

def main():
    engine = get_engine()
    engine.connect()
    try:
        doctest.testmod(verbose=True)
    finally:
        engine.disconnect()

if __name__ == "__main__":
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Test cases for the logging framework
============================================================================

"""

import sys
import logging
import logging.handlers
import unittest
import dragonfly.log as log


#===========================================================================

class OutputCapturer(object):

    def __init__(self):
        self.blocks = []

    def write(self, data):
        self.blocks.append(data)

    def flush(self):
        pass

    def clear(self):
        self.blocks = []

    @property
    def lines(self, prefix=""):
        if not self.blocks:
            return ()
        else:
            text = "".join(self.blocks).splitlines()
            text = prefix + ("\n" + prefix).join(text)
            return text.splitlines()


#---------------------------------------------------------------------------

class LogTestCase(unittest.TestCase):
    """ Test behavior of logging system. """

    def setUp(self):
        self._original_stdout = sys.stdout
        self._output = OutputCapturer()
        sys.stdout = self._output
        self._original_stderr = sys.stderr
        self._error = OutputCapturer()
        sys.stderr = self._error

    def tearDown(self):
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr
#        if self._output.blocks:
#            prefix = "Output:  "
#            output = "".join(self._output.blocks).splitlines()
#            output = prefix + ("\n" + prefix).join(output)
#            print output
#        if self._error.blocks:
#            prefix = "Error:   "
#            text = "".join(self._error.blocks).splitlines()
#            text = prefix + ("\n" + prefix).join(text)
#            print text
        self._output = None
        self._error = None

    def test_filtering(self):
        """ Verify that log messages are filtered according to level. """
        log.setup_log()
        logger = logging.getLogger("grammar")
        logger.debug("test_filtering - debug")
        logger.info("test_filtering - info")
        logger.warning("test_filtering - warning")
        logger.error("test_filtering - error")
        expected = ["grammar (WARNING): test_filtering - warning",
                    "grammar (ERROR): test_filtering - error"]
        self.assertEqual(self._error.lines, expected)

        self._error.clear()
        logger = logging.getLogger("grammar.begin")
        logger.debug("test_filtering - debug")
        logger.info("test_filtering - info")
        logger.warning("test_filtering - warning")
        logger.error("test_filtering - error")
        expected = ["grammar.begin (INFO): test_filtering - info",
                    "grammar.begin (WARNING): test_filtering - warning",
                    "grammar.begin (ERROR): test_filtering - error"]
        self.assertEqual(self._error.lines, expected)

        self._error.clear()
        logger = logging.getLogger("grammar.load")
        logger.debug("test_filtering - debug")
        logger.info("test_filtering - info")
        logger.warning("test_filtering - warning")
        logger.error("test_filtering - error")
        expected = ["grammar.load (WARNING): test_filtering - warning",
                    "grammar.load (ERROR): test_filtering - error"]
        self.assertEqual(self._error.lines, expected)

    def _new_lines(self):
        filename = None
        if not hasattr(self, "_previous_line_count"):
            self._previous_line_count = 0
        lines = open(filename).readlines()
        new_lines = lines[self._previous_line_count:]
        self._previous_line_count = len(lines)
        return new_lines


#===========================================================================

if __name__ == "__main__":
    unittest.main()



from time import sleep

from dragonfly import PlaySound

# See for more sound names:
# http://www.python.org/doc/2.5.2/lib/module-winsound.html
PlaySound("SystemHand").execute()
sleep(1)
PlaySound(file=r"C:\WINDOWS\Media\Windows XP Battery Low.wav").execute()
sleep(1)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
"""

import logging
from six import string_types

from dragonfly              import *
from .error                 import TestError
from ..test                 import infrastructure


#===========================================================================

class RuleTestGrammar(Grammar):

    _log = logging.getLogger("test")
    _NoRecognition = infrastructure.Unique("NoRecognition")

    #-----------------------------------------------------------------------

    def __init__(self, name=None, engine=None, context=None):
        if name is None:
            name = self.__class__.__name__
        Grammar.__init__(self, name=name, engine=engine, context=context)

    def recognize(self, words):
        if isinstance(words, string_types):
            words = words.split()

        if not self.loaded:
            self._log.debug("Loading TesterGrammar.")
            self.load()
            unload_after_recognition = True
        else:
            unload_after_recognition = False

        try:
            # Make this grammar exclusive; this *probably* avoids other
            #  grammars from being active and receiving the mimicked
            #  recognition.
            try:
                self.set_exclusiveness(True)
            except Exception as e:
                msg = ("Exception during setting grammar as"
                       " exclusive: %s" % (e,))
                self._log.exception(msg)
                raise TestError(msg)

            # Mimic recognition.
            try:
                self.engine.mimic(words)
            except MimicFailure as e:
                msg = "Recognition failed. (Words: %s)" % (words,)
                self._log.error(msg)
                raise TestError(msg)
            except Exception as e:
                msg = ("Exception during recognition: %s" % (e,))
                self._log.exception(msg)
                raise TestError(msg)

        finally:
            if unload_after_recognition:
                try:
                    self.unload()
                except Exception as e:
                    # Log exception, but do not reraise it as that would
                    #  mask any earlier exception.
                    self._log.exception("Failed to unload grammar: %s"
                                        % (e,))

    def recognize_node(self, words):
        """
            Mimic recognition of the given words and return the
            root node of the recognized rule.

            The root node of the recognized rule is collected by
            patching the `process_recognition` methods of all
            loaded rules in this grammar with a custom function.
            Note that this patching prevents any of the rules'
            original recognition processing logic from being
            executed.

        """
        # Logic to patch rules with process recognition for testing.
        self._recognized_node = self._NoRecognition
        def test_process_recognition(node):
            self._recognized_node = node

        # Patch rules with process recognition method for testing.
        patched_rules = []
        for rule in self.rules:
            if hasattr(rule, "process_recognition"):
                original_method = rule.process_recognition
                rule.process_recognition = test_process_recognition
                patched_rules.append((rule, original_method))

        try:
            self.recognize(words)

            # If no node was collected, then none of the
            #  patch rules received the recognition.
            if self._recognized_node == self._NoRecognition:
                msg = "Recognition hijacked. (Words: %s)" % (words,)
                self._log.error(msg)
                raise TestError(msg)
            return self._recognized_node

        finally:
            # Restore original process recognition methods.
            for rule, original_method in patched_rules:
                rule.process_recognition = original_method


    def recognize_extras(self, words):
        """
            Mimic recognition of the given words and return the
            the extras parameter for the recognized rule.

            The extras parameter for the recognized rule is collected
            by patching the `_process_recognition` methods of all
            loaded rules in this grammar with a custom function.
            Note that this patching prevents any of the rules'
            original recognition processing logic from being
            executed.

        """
        # Logic to patch rules with process recognition for testing.
        self._recognized_node = self._NoRecognition
        self._recognized_extras = self._NoRecognition
        def test_process_recognition(node, extras):
            self._recognized_node = node
            self._recognized_extras = extras

        # Patch rules with process recognition method for testing.
        patched_rules = []
        for rule in self.rules:
            if hasattr(rule, "_process_recognition"):
                original_method = rule._process_recognition
                rule._process_recognition = test_process_recognition
                patched_rules.append((rule, original_method))

        try:
            self.recognize(words)

            # If no extras were collected, then none of the
            #  patch rules received the recognition.
            if self._recognized_extras == self._NoRecognition:
                msg = "Recognition hijacked. (Words: %s)" % (words,)
                self._log.error(msg)
                raise TestError(msg)
            return self._recognized_extras

        finally:
            # Restore original process recognition methods.
            for rule, original_method in patched_rules:
                rule._process_recognition = original_method



import unittest

import time
import logging

from dragonfly import (CompoundRule, Function, Grammar, MappingRule)
from dragonfly.engines import (EngineBase, EngineError, MimicFailure,
                               get_engine)
from dragonfly.grammar.elements import Dictation, Literal, Sequence
from dragonfly.grammar.recobs import RecognitionObserver
from dragonfly.test import (ElementTester, RuleTestGrammar)
from dragonfly.test.test_engine_sphinx import SphinxEngineCase


# class DictationEngineTests(SphinxEngineCase):
class DictationEngineTests(object):
    """
    Tests for the engine's dictation functionality.

    Dictation andthese tests are temporarily **DISABLED** to make other
    changes easier.
    """
    def test_single_dictation(self):
        """
        Test that the engine can handle a dragonfly rule using a Dictation element.
        """
        test = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {"<dictation>": Function(test)}
            extras = [Dictation("dictation")]

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test, 1)

        # Test that it works again
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test, 2)

        # Test again with multiple words
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test, 3)

    def test_dictation_with_other_elements(self):
        """
        Test that the engine can handle dragonfly rules using dictation with other
        elements.

        This is being tested because the way dictation works for this engine is
        different than how the other engines handle it: Sphinx needs utterance
        pauses between grammar and dictation rule parts.
        """
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "hello <dictation>": Function(test1),
                "<dictation> testing": Function(test2),
            }
            extras = [Dictation("dictation")]

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)

        # Dictation and other elements are processed separately with pauses
        self.assert_mimic_success("hello", "world")  # test mapping 1
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)

        # Test with mapping 2
        self.assert_mimic_success("testing", "testing")

        # Test that it works again with mapping 1
        self.reset_test_functions()
        self.assert_mimic_success("hello", "world")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)

        # Test with mapping 2 again
        self.reset_test_functions()
        self.assert_mimic_success("test", "testing")
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 1)

        # Test again with multiple words
        self.reset_test_functions()
        self.assert_mimic_success("hello", "hi there")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)

        # And test again using mapping 2
        self.reset_test_functions()
        self.assert_mimic_success("test test test", "testing")
        self.assert_test_function_called(test2, 1)
        self.assert_test_function_called(test1, 0)

        # Test incomplete sequences
        self.reset_test_functions()
        self.assert_mimic_success("hello")  # start of mapping one
        self.assert_test_function_called(test1, 0)  # incomplete sequence
        self.assert_mimic_success("test", "testing")
        self.assert_test_function_called(test1, 1)  # mapping one completely matched
        # mapping two only partially matched with <dictation> as "testing"
        self.assert_test_function_called(test2, 0)

    def test_recognition_observer(self):
        """
        Test that the engine's recognition observer manager works correctly.
        """
        on_begin_test = self.get_test_function()
        on_recognition_test = self.get_test_function()
        on_failure_test = self.get_test_function()
        on_next_rule_part_test = self.get_test_function()

        # Set up a custom observer using test methods
        class TestObserver(RecognitionObserver):
            on_begin = on_begin_test
            on_next_rule_part = on_next_rule_part_test
            on_recognition = on_recognition_test
            on_failure = on_failure_test

        # Set up a TestObserver instance and a grammar with multiple rules to use
        observer = TestObserver()
        observer.register()
        grammar = Grammar("test")
        grammar.add_rule(CompoundRule("rule1", "hello world"))
        grammar.add_rule(CompoundRule("rule2", "say <dictation>",
                                      extras=[Dictation("dictation")]))
        grammar.load()
        self.assertTrue(grammar.loaded)

        # Test that each method is called properly
        self.assert_mimic_success("hello world")

        # on_begin is called during each mimic. on_recognition should be called
        # once per successful and complete recognition. Both on_failure and
        # on_next_rule_part shouldn't have been called yet.
        self.assert_test_function_called(on_begin_test, 1)
        self.assert_test_function_called(on_recognition_test, 1)
        self.assert_test_function_called(on_failure_test, 0)
        self.assert_test_function_called(on_next_rule_part_test, 0)

        # Test with a dictation rule
        self.assert_mimic_success("say")

        # Recognition begins again, is incomplete and no failure yet.
        self.assert_test_function_called(on_begin_test, 2)
        self.assert_test_function_called(on_recognition_test, 1)
        self.assert_test_function_called(on_failure_test, 0)

        # on_next_rule_part should be called because there are more rule parts
        self.assert_test_function_called(on_next_rule_part_test, 1)

        # Test the next part of the dictation rule
        self.assert_mimic_success("testing testing")

        # Recognition begins again, is complete, and no failure yet.
        self.assert_test_function_called(on_begin_test, 3)
        self.assert_test_function_called(on_recognition_test, 2)
        self.assert_test_function_called(on_failure_test, 0)

        # on_next_rule_part shouldn't have been called because this is the last part
        # and on_recognition will be called instead
        self.assert_test_function_called(on_next_rule_part_test, 1)

        # Recognition begins again and failure occurs.
        self.assert_mimic_failure("testing")
        self.assert_test_function_called(on_begin_test, 4)
        self.assert_test_function_called(on_next_rule_part_test, 1)  # no change
        self.assert_test_function_called(on_failure_test, 1)
        self.assert_test_function_called(on_recognition_test, 2)

        # Test that using None or "" also calls the on_failure method
        self.assert_mimic_failure(None)
        self.assert_test_function_called(on_failure_test, 2)
        self.assert_mimic_failure("")
        self.assert_test_function_called(on_failure_test, 3)
        self.assert_test_function_called(on_next_rule_part_test, 1)  # no change

        # Unregister the observer
        observer.unregister()

    def test_no_hypothesis(self):
        """
        Check that if something that doesn't match any rule is mimicked, nothing
        gets recognised.
        """
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "testing": Function(test1),
                "<dictation>": Function(test2)
            }
            extras = [Dictation("dictation")]

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)

        self.assert_mimic_failure(None)
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 0)
        self.assert_mimic_failure("")
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 0)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 1)

    def test_sequence_timeout(self):
        """
        Test that when speaking subsequent parts of rules involving Dictation
        elements, the recognition times out if there is a timeout period set.
        """
        # Set a timeout period for this test of 100 ms, which is unreasonable for
        # humans, but fine for mimic().
        timeout = 0.1
        self.engine.config.NEXT_PART_TIMEOUT = timeout

        # Set up a test grammar with some rules
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "say <dictation>": Function(test1),
                "hello world": Function(test2),
            }
            extras = [Dictation("dictation")]

        grammar = Grammar("test")
        grammar.add_rule(TestRule())
        grammar.load()
        self.assertTrue(grammar.loaded)

        # Test that mapping 1 can be recognised fully
        self.assert_mimic_success("say", "hello")
        self.assert_test_function_called(test1, 1)

        # Start recognising mapping 1 again, then sleep for a bit
        self.assert_mimic_success("say")
        time.sleep(timeout)

        # The rest of mapping 1 should not match
        self.assert_mimic_failure("testing")
        self.assert_test_function_called(test1, 1)  # still only called 1 time

        # Should be able to match mapping 2 now
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test2, 1)

        # Test that it works with a break shorter than the timeout value
        self.assert_mimic_success("say")
        time.sleep(timeout / 2)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test1, 2)

        # Test with a timeout of 0 (no timeout)
        timeout = 0
        self.engine.config.NEXT_PART_TIMEOUT = timeout
        self.assert_mimic_success("say")
        time.sleep(0.1)  # sleep for 100ms
        self.assert_test_function_called(test1, 2)  # no change
        self.assert_mimic_success("testing")
        self.assert_test_function_called(test1, 3)

        # Test without a sleep between rule parts
        self.assert_mimic_success("say", "testing")
        self.assert_test_function_called(test1, 4)

        # Test that mapping 2 still has no issue matching
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test2, 2)

    def test_long_sequence_rule(self):
        """
        Test that the engine can recognise a long sequence rule and that the
        timeout functionality works correctly.
        """
        timeout = 0.1
        self.engine.config.NEXT_PART_TIMEOUT = timeout

        # Set up a test grammar with some rules
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "testing": Function(test1),
                "testing <dictation> testing <dictation>": Function(test2),
            }
            extras = [Dictation("dictation")]

        grammar = Grammar("test")
        grammar.add_rule(TestRule())
        grammar.load()
        self.assertTrue(grammar.loaded)

        # Test that mapping 2 can be recognised fully
        self.assert_mimic_success("testing", "hello", "testing")
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test2, 1)
        self.assert_test_function_called(test1, 0)

        # Start recognising mapping 2 again, then sleep for a bit
        self.assert_mimic_success("testing")
        time.sleep(timeout + 0.1)

        # Only mapping 1 should have been processed
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 1)

        # Test that it works with a break shorter than the timeout value
        self.assert_mimic_success("testing")
        time.sleep(timeout / 2)
        self.assert_mimic_success("hello")
        time.sleep(timeout / 2)
        self.assert_mimic_success("testing")
        time.sleep(timeout / 2)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test2, 2)

        # Test with a timeout of 0 (no timeout)
        timeout = 0
        self.engine.config.NEXT_PART_TIMEOUT = timeout
        self.assert_mimic_success("testing")
        time.sleep(0.1)  # sleep for 100ms
        self.assert_test_function_called(test2, 2)  # no changes

        # Test that mapping 1 won't process any more, even after a time.
        self.assert_test_function_called(test1, 1)

        # Mimicking the rest of the rule parts should make the rule process.
        self.assert_mimic_success("hello", "testing", "hello")
        self.assert_test_function_called(test2, 3)



# encoding=utf-8
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import os
from tempfile import NamedTemporaryFile, mkdtemp
import unittest

from dragonfly import Clipboard


format_unicode = Clipboard.format_unicode
format_text = Clipboard.format_text
format_hdrop = Clipboard.format_hdrop


class TestClipboard(unittest.TestCase):
    """
    Tests for the Clipboard class.

    These should pass on all supported platforms.
    """

    #-----------------------------------------------------------------------

    @classmethod
    def setUpClass(cls):
        # Save the current contents of the clipboard.
        cls.clipboard = Clipboard(from_system=True)

    @classmethod
    def tearDownClass(cls):
        # Restore the clipboard contents to what it was before the tests.
        cls.clipboard.copy_to_system()

    def setUp(self):
        # Clear the clipboard before each test.
        Clipboard.clear_clipboard()

    #-----------------------------------------------------------------------

    def test_get_set_system_text(self):
        # Test with an empty system clipboard.
        Clipboard.clear_clipboard()
        self.assertEqual(Clipboard.get_system_text(), u"")

        # Test setting a Unicode string on the system clipboard.
        text1 = u"Unicode text"
        Clipboard.set_system_text(text1)
        self.assertEqual(Clipboard.get_system_text(), text1)

        # Test setting a binary string on the system clipboard.
        # get_system_text() should return the equivalent Unicode string, as
        # format_unicode is preferred.
        Clipboard.set_system_text(b"text")
        self.assertEqual(Clipboard.get_system_text(), u"text")

        # Test that setting the text using None clears the clipboard.
        Clipboard.set_system_text(None)
        self.assertEqual(Clipboard.get_system_text(), u"")

    def test_clear_clipboard(self):
        # Put something on the system clipboard.
        Clipboard.set_system_text("something")

        # Clear it.
        Clipboard.clear_clipboard()

        # Then test that it has been cleared.
        self.assertEqual(Clipboard.get_system_text(), "")

    def test_convert_format_content(self):
        # Define a convenient shorthand function for testing the
        # Clipboard.convert_format_content() class method.
        def convert(format, content):  # pylint: disable=redefined-builtin
            return Clipboard.convert_format_content(format, content)

        # Text strings used with format_text (ANSI) are converted.
        self.assertEqual(convert(format_text, u"text"), b"text")

        # Text strings used with format_unicode are not changed.
        self.assertEqual(convert(format_unicode, u"text"), u"text")

        # Binary strings used with format_unicode are converted.
        self.assertEqual(convert(format_unicode, b"text"), u"text")

        # Binary strings used with format_text are not changed.
        self.assertEqual(convert(format_text, b"text"), b"text")

        # Test format_hdrop with existing file paths.
        with NamedTemporaryFile() as f1, NamedTemporaryFile() as f2:
            # Strings are converted to tuples containing file paths. Test
            # using Unicode and binary strings.
            # First, test using one existing file.
            for test_string in [u"%s\0\0", u"%s\n\n", u"%s\0", u"%s\n"]:
                test_string = test_string % f1.name
                f1name = u"%s" % f1.name
                self.assertEqual(convert(format_hdrop, test_string),
                                 (f1name,))
                test_string = test_string.encode()
                f1name = f1name.encode()
                self.assertEqual(convert(format_hdrop, test_string),
                                 (f1name,))

            # Then, test using two.
            for test_string in ["%s\0%s\0\0", "%s\n%s\0", "%s\n%s\n"]:
                test_string = test_string % (f1.name, f2.name)
                f1name, f2name = u"%s" % f1.name, u"%s" % f2.name
                self.assertEqual(convert(format_hdrop, test_string),
                                 (f1name, f2name))
                test_string = test_string.encode()
                f1name, f2name = f1name.encode(), f2name.encode()
                self.assertEqual(convert(format_hdrop, test_string),
                                 (f1name, f2name))

            # Tuples and lists containing strings are accepted. Lists are
            # converted to tuples.
            self.assertEqual(convert(format_hdrop, (f1.name,)), (f1.name,))
            self.assertEqual(convert(format_hdrop, [f1.name]), (f1.name,))

        # Existing relative file paths are not accepted.
        cwd = os.getcwd()
        try:
            # Make a temporary directory and enter it.
            tempdir = mkdtemp()
            os.chdir(tempdir)

            # Test the relative file path.
            with NamedTemporaryFile(dir=tempdir) as f:
                basename = u"%s" % os.path.basename(f.name)
                for test_obj in [basename, basename.encode(), [basename],
                                 (basename,)]:
                    self.assertRaises(ValueError, convert, format_hdrop,
                                      test_obj)
        finally:
            # Return to the original working directory and delete tempdir.
            os.chdir(cwd)
            os.rmdir(tempdir)

        # Errors are raised when unacceptable content is used with the text
        # formats or with format_hdrop.
        for test_obj in [object(), None, 0]:
            self.assertRaises(TypeError, convert, format_text, test_obj)
            self.assertRaises(TypeError, convert, format_unicode, test_obj)
            self.assertRaises(TypeError, convert, format_hdrop, test_obj)
            self.assertRaises(TypeError, convert, format_hdrop, (test_obj,))
            self.assertRaises(TypeError, convert, format_hdrop, [test_obj])
        for test_string in [u"", u"\n", u"\0", u"test\n", u"test\0"]:
            self.assertRaises(ValueError, convert, format_hdrop,
                              test_string)
            self.assertRaises(ValueError, convert, format_hdrop,
                              test_string.encode())

        # Empty lists/tuples are not acceptable format_hdrop content.
        self.assertRaises(ValueError, convert, format_hdrop, [])
        self.assertRaises(ValueError, convert, format_hdrop, ())

    def test_from_system_argument(self):
        # Test the optional from_system argument of Clipboard.__init__
        text = "something"
        Clipboard.set_system_text(text)
        c = Clipboard(from_system=True)
        self.assertEqual(c.text, text)
        self.assertTrue(c.has_text())

    def test_copy_from_system(self):
        text = "testing"
        Clipboard.set_system_text(text)
        c = Clipboard()

        # Test the method with clear=False (default).
        c.copy_from_system(clear=False)
        self.assertEqual(c.text, text)
        self.assertEqual(Clipboard.get_system_text(), text)

        # Test again with clear=True.
        c = Clipboard()
        c.copy_from_system(clear=True)
        self.assertEqual(c.text, text)
        self.assertEqual(Clipboard.get_system_text(), "")

        # Test formats=format_unicode.
        # Set the system clipboard before testing.
        text1 = u"unicode text"
        c1 = Clipboard(contents={format_unicode: text1})
        c1.copy_to_system()
        c2 = Clipboard()
        c2.copy_from_system(formats=format_unicode)
        self.assertTrue(c2.has_format(format_unicode))
        self.assertEqual(c2.get_format(format_unicode), text1)

        # Test formats=format_text.
        # Set the system clipboard before testing.
        text2 = b"text"
        c1 = Clipboard(contents={format_text: text2})
        c1.copy_to_system()
        c2 = Clipboard()
        c2.copy_from_system(formats=format_text)
        self.assertTrue(c2.has_format(format_text))
        self.assertEqual(c2.get_format(format_text), text2)

        # Test formats=(format_unicode, format_text).
        # Set the system clipboard before testing. Use the same string for
        # both formats so the test will work on all platforms.
        c1 = Clipboard(contents={format_text: b"text",
                                 format_unicode: u"text"})
        c1.copy_to_system()
        c2 = Clipboard()
        c2.copy_from_system(formats=(format_unicode, format_text))
        self.assertTrue(c2.has_format(format_unicode))
        self.assertEqual(c2.get_format(format_unicode), u"text")
        self.assertTrue(c2.has_format(format_text))
        self.assertEqual(c2.get_format(format_text), b"text")

    def test_copy_to_system(self):
        # Test with format_unicode.
        text1 = u"unicode text"
        c = Clipboard(contents={format_unicode: text1})
        c.copy_to_system(clear=True)
        self.assertEqual(Clipboard.get_system_text(), text1)

        # Test with format_text. Text string used deliberately here;
        # get_system_text() should returns those.
        text2 = u"text"
        c = Clipboard(contents={format_text: text2})
        c.copy_to_system(clear=True)
        self.assertEqual(Clipboard.get_system_text(), text2)

        # Test with text.
        text3 = u"testing"
        c = Clipboard(text=text3)
        c.copy_to_system(clear=True)
        self.assertEqual(Clipboard.get_system_text(), text3)

        # Test with an empty Clipboard instance.
        c = Clipboard()
        c.copy_to_system(clear=False)
        self.assertEqual(Clipboard.get_system_text(), text3)
        c.copy_to_system(clear=True)
        self.assertEqual(Clipboard.get_system_text(), u"")

    def test_get_available_formats(self):
        # Test with an empty Clipboard instance.
        c = Clipboard()
        self.assertEqual(c.get_available_formats(), [])

        # Test with one format.
        c = Clipboard(contents={format_unicode: u"unicode text"})
        self.assertEqual(c.get_available_formats(), [format_unicode])

        # Test with two formats.
        c = Clipboard(contents={format_unicode: u"unicode text",
                                format_text: u"text"})
        self.assertEqual(c.get_available_formats(), [format_unicode,
                                                     format_text],
                         "format_unicode should precede format_text")

        # Test with text formats and one custom format. The preferred text
        # format should be first on the list.
        format_custom = 0
        c = Clipboard(contents={format_custom: 123,
                                format_unicode: u"unicode text"})
        self.assertEqual(c.get_available_formats(), [format_unicode,
                                                     format_custom],
                         "format_unicode should precede format_custom")
        c = Clipboard(contents={format_custom: 123,
                                format_text: b"text"})
        self.assertEqual(c.get_available_formats(), [format_text,
                                                     format_custom],
                         "format_text should precede format_custom")

        # Test with no text format and one custom format.
        c = Clipboard(contents={format_custom: 123})
        self.assertEqual(c.get_available_formats(), [format_custom])

    def test_has_format(self):
        # Test with an empty Clipboard instance.
        c = Clipboard()
        self.assertFalse(c.has_format(format_unicode))
        self.assertFalse(c.has_format(format_text))
        self.assertFalse(c.has_format(format_hdrop))

        # Test with one format.
        c = Clipboard(contents={format_unicode: u"unicode text"})
        self.assertTrue(c.has_format(format_unicode))
        self.assertFalse(c.has_format(format_text))

        # Test with two formats.
        c = Clipboard(contents={format_unicode: u"unicode text",
                                format_text: u"text"})
        self.assertTrue(c.has_format(format_unicode))
        self.assertTrue(c.has_format(format_text))

    def test_get_format(self):
        # Test with an empty Clipboard instance.
        c = Clipboard()
        self.assertRaises(ValueError, c.get_format, format_unicode)
        self.assertRaises(ValueError, c.get_format, format_text)
        self.assertRaises(ValueError, c.get_format, format_hdrop)

        # Test with one format.
        text1 = u"unicode text"
        c = Clipboard(contents={format_unicode: text1})
        self.assertEqual(c.get_format(format_unicode), text1)
        self.assertRaises(ValueError, c.get_format, format_text)
        self.assertRaises(ValueError, c.get_format, format_hdrop)

        # Test with two formats.
        text2 = b"text"
        c = Clipboard(contents={format_unicode: text1,
                                format_text: text2})
        self.assertEqual(c.get_format(format_unicode), text1)
        self.assertEqual(c.get_format(format_text), text2)
        self.assertRaises(ValueError, c.get_format, format_hdrop)

        # Test with format_hdrop.
        with NamedTemporaryFile() as f:
            c = Clipboard(contents={format_hdrop: f.name})
            self.assertRaises(ValueError, c.get_format, format_unicode)
            self.assertRaises(ValueError, c.get_format, format_text)
            self.assertEqual(c.get_format(format_hdrop), (f.name,))

    def test_set_format(self):
        # Test with one format.
        text1 = u"unicode text"
        c = Clipboard()
        c.set_format(format_unicode, text1)
        self.assertTrue(c.has_format(format_unicode))
        self.assertEqual(c.get_format(format_unicode), text1)
        self.assertFalse(c.has_format(format_text))
        self.assertRaises(ValueError, c.get_format, format_text)

        # Test with two formats.
        text2 = b"text"
        c.set_format(format_text, text2)
        self.assertTrue(c.has_format(format_unicode))
        self.assertEqual(c.get_format(format_unicode), text1)
        self.assertTrue(c.has_format(format_text))
        self.assertEqual(c.get_format(format_text), text2)

        # Test with format_hdrop.
        with NamedTemporaryFile() as f:
            c.set_format(format_hdrop, f.name)
            self.assertTrue(c.has_format(format_unicode))
            self.assertTrue(c.has_format(format_text))
            self.assertTrue(c.has_format(format_hdrop))
            self.assertEqual(c.get_format(format_hdrop), (f.name,))

        # Setting a format to None removes the format's content from the
        # instance.
        c.set_format(format_text, None)
        self.assertFalse(c.has_format(format_text))
        self.assertRaises(ValueError, c.get_format, format_text)
        self.assertTrue(c.has_format(format_unicode))
        self.assertTrue(c.has_format(format_hdrop))
        c.set_format(format_unicode, None)
        self.assertFalse(c.has_format(format_unicode))
        self.assertRaises(ValueError, c.get_format, format_unicode)
        self.assertTrue(c.has_format(format_hdrop))
        c.set_format(format_hdrop, None)
        self.assertFalse(c.has_format(format_hdrop))
        self.assertRaises(ValueError, c.get_format, format_hdrop)

    def test_has_text(self):
        # Test with an empty Clipboard instance.
        c = Clipboard()
        self.assertFalse(c.has_text())

        # Test with format_unicode only.
        c = Clipboard(contents={format_unicode: u"unicode text"})
        self.assertTrue(c.has_text())

        # Test with both text formats.
        c = Clipboard(contents={format_unicode: u"unicode text",
                                format_text: b"text"})
        self.assertTrue(c.has_text())

        # Test with format_text only.
        c = Clipboard(contents={format_text: b"text"})
        self.assertTrue(c.has_text())

        # Test with format_hdrop only.
        with NamedTemporaryFile() as f:
            c = Clipboard(contents={format_hdrop: f.name})
            self.assertFalse(c.has_text())

    def test_set_text(self):
        c = Clipboard()
        text = "test"
        c.set_text(text)
        self.assertTrue(c.has_text())
        self.assertEqual(c.get_text(), text)
        self.assertEqual(c.text, text)

        # Setting the text to None clears the stored text.
        c.set_text(None)
        self.assertFalse(c.has_text())
        self.assertIsNone(c.get_text())

        # Test setting the text using the text property.
        c.text = text
        self.assertTrue(c.has_text())
        self.assertEqual(c.get_text(), text)
        self.assertEqual(c.text, text)

    def test_get_text(self):
        # Test with an empty Clipboard instance.
        c = Clipboard()
        self.assertIsNone(c.get_text())

        # Test with Unicode text.
        text1 = u"test"
        c.set_text(text1)
        self.assertEqual(c.get_text(), text1)

        # Test with text set to None.
        c.set_text(None)
        self.assertIsNone(c.get_text())

        # Test with binary text.
        text2 = b"test"
        c.set_text(text2)
        self.assertEqual(c.get_text(), text2)

        # Test with format_hdrop.
        with NamedTemporaryFile() as f:
            c = Clipboard(contents={format_hdrop: f.name})
            self.assertIsNone(c.get_text())

    def test_comparison(self):
        # Test with two empty Clipboard instances.
        self.assertEqual(Clipboard(), Clipboard())

        # Test with format_unicode.
        c1 = Clipboard(contents={format_unicode: u"unicode text"})
        c2 = Clipboard(contents={format_unicode: u"unicode text"})
        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, Clipboard())

        # Test with the same clipboard instance.
        self.assertEqual(c1, c1)

        # Test with both text formats.
        c1 = Clipboard(contents={format_unicode: u"unicode text",
                                 format_text: b"text"})
        c2 = Clipboard(contents={format_unicode: u"unicode text",
                                 format_text: b"text"})
        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, Clipboard())

        # Test with format_text only.
        c1 = Clipboard(contents={format_text: b"text"})
        c2 = Clipboard(contents={format_text: b"text"})
        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, Clipboard())

        # Test with different text formats.
        c1 = Clipboard(contents={format_unicode: u"unicode text"})
        c2 = Clipboard(contents={format_text: b"text"})
        self.assertNotEqual(c1, c2)

        # Test with format_hdrop only.
        with NamedTemporaryFile() as f:
            c1 = Clipboard(contents={format_hdrop: f.name})
            c2 = Clipboard(contents={format_hdrop: f.name})
            self.assertEqual(c1, c2)
            self.assertNotEqual(c1, Clipboard())
            self.assertNotEqual(c1, Clipboard(text="text"))

    def test_flexible_string_types(self):
        # This is similar to the clipboard format conversion that Windows
        # performs when necessary. The Clipboard class should do this
        # regardless of platform/implementation.

        # Binary strings used with format_unicode are converted for us.
        c = Clipboard(contents={format_unicode: b"text"})
        self.assertEqual(c.get_format(format_unicode), u"text")
        c = Clipboard()
        c.set_format(format_unicode, b"text")
        self.assertEqual(c.get_format(format_unicode), u"text")

        # Text strings used with format_text (ANSI) are converted for us.
        c = Clipboard(contents={format_text: u"text"})
        self.assertEqual(c.get_format(format_text), b"text")
        c = Clipboard()
        c.set_format(format_text, u"text")
        self.assertEqual(c.get_format(format_text), b"text")

    def test_non_ascii(self):
        text = u"""
        ૱ ꠸ ┯ ┰ ┱ ┲ ❗ ► ◄ Ă ă 0 1 2 3 4 5 6 7 8 9 Ǖ ǖ Ꞁ ¤ ­ Ð ¢ ℥ Ω ℧ K ℶ
        ℷ ℸ ⅇ ⅊ ⚌ ⚍ ⚎ ⚏ ⚭ ⚮ ⌀ ⏑ ⏒ ⏓ ⏔ ⏕ ⏖ ⏗ ⏘ ⏙ ⏠ ⏡ ⏦ ᶀ ᶁ ᶂ ᶃ ᶄ ᶆ ᶇ ᶈ
        ᶉ ᶊ ᶋ ᶌ ᶍ ᶎ ᶏ ᶐ ᶑ ᶒ ᶓ ᶔ ᶕ ᶖ ᶗ ᶘ ᶙ ᶚ ᶸ ᵯ ᵰ ᵴ ᵶ ᵹ ᵼ ᵽ ᵾ ᵿ
        """
        Clipboard.set_system_text(text)
        self.assertEqual(Clipboard.get_system_text(), text)

    def test_non_strings(self):
        # Using non-string objects for text formats raises errors.
        self.assertRaises(TypeError, Clipboard, {format_text: 0})
        self.assertRaises(TypeError, Clipboard, {format_unicode: 0})
        self.assertRaises(TypeError, Clipboard, {format_text: object()})
        self.assertRaises(TypeError, Clipboard, {format_unicode: object()})
        self.assertRaises(TypeError, Clipboard, {format_text: None})
        self.assertRaises(TypeError, Clipboard, {format_unicode: None})

        c = Clipboard()
        self.assertRaises(TypeError, c.set_text, 0)
        self.assertRaises(TypeError, c.set_text, object())
        self.assertRaises(TypeError, c.set_format, format_text, 0)
        self.assertRaises(TypeError, c.set_format, format_text, object())
        self.assertRaises(TypeError, c.set_format, format_unicode, 0)
        self.assertRaises(TypeError, c.set_format, format_unicode, object())
        self.assertRaises(TypeError, c.set_system_text, 0)
        self.assertRaises(TypeError, c.set_system_text, object())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#

import sys
import os
import os.path
import re

directory = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0, directory)

#---------------------------------------------------------------------------
# Set an environment variable so that dragonfly can check if a sphinx build
# is running.

os.environ["SPHINX_BUILD_RUNNING"] = "1"


#---------------------------------------------------------------------------
# Gather version from distribution file.

path = os.path.join(directory, "version.txt")
version_string = open(path).readline()
match = re.match(r"\s*(?P<rel>(?P<ver>\d+\.\d+)(?:\.\S+)*)\s*", version_string)
version = match.group("ver")
release = match.group("rel")
print("Version:", version, "-- Release:", release)


#---------------------------------------------------------------------------
# Mock libraries that are only available on some platforms or with optional
# dependencies installed.

from mock import MagicMock

class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()


mock_modules = {
    # Modules required on Windows
    "ctypes.wintypes",
    "pythoncom",
    "pywintypes",
    "win32api",
    "win32clipboard",
    "win32com",
    "win32com.client",
    "win32com.client.gencache",
    "win32com.gen_py",
    "win32com.shell",
    "win32con",
    "win32event",
    "win32file",
    "win32gui",
    "winsound",
    "winxpgui",

    # Modules required on Linux
    "psutil",

    # Modules required on macOS
    "AppKit",
    "applescript",

    # Modules required by dragonfly.rpc
    "decorator",
    "jsonrpc",
    "jsonrpc.dispatcher",
    "jsonrpc.manager",
    "werkzeug",
    "werkzeug.serving",
    "werkzeug.wrappers",

    # Modules required by Kaldi engine
    "kaldi_active_grammar",
    "sounddevice",
    "webrtcvad",

    # Modules required by Sphinx engine
    "jsgf",
    "jsgf.ext",
    "pocketsphinx",
    "pyaudio",
    "sphinxwrapper",

    # Other
    "ctypes",
    "numpy",
    "pyperclip",
    "regex",
}

for module_name in mock_modules:
    sys.modules[module_name] = Mock()


#---------------------------------------------------------------------------
# General configuration

extensions = ["sphinx.ext.autodoc", "sphinx.ext.viewcode", "sphinxarg.ext"]
templates_path = ["templates"]
source_suffix = ".txt"
master_doc = "index"

project = u"Dragonfly"
copyright = u"2014, Christo Butcher"

today_fmt = "%Y-%m-%d"
exclude_patterns = ["requirements.txt"]
add_module_names = False
pygments_style = "sphinx"
autoclass_content = "both"


#---------------------------------------------------------------------------
# Options for HTML output

html_theme = "sphinx_rtd_theme"
html_last_updated_fmt = "%Y-%m-%d"
html_copy_source = True
htmlhelp_basename = "Dragonflydoc"


#---------------------------------------------------------------------------
# Options for LaTeX output

latex_documents = [
  ("index", "Dragonfly.tex", u"Dragonfly Documentation",
   u"Christo Butcher", "manual"),
]



from __future__ import print_function
import logging, os

import dragonfly

if False:
    logging.basicConfig(level=10)
    logging.getLogger('grammar.decode').setLevel(20)
    logging.getLogger('compound').setLevel(20)
    # logging.getLogger('kaldi').setLevel(30)
    logging.getLogger('engine').setLevel(10)
    logging.getLogger('kaldi').setLevel(10)
else:
    # logging.basicConfig(level=20)
    from dragonfly.log import setup_log
    setup_log()


class ExampleCustomRule(dragonfly.CompoundRule):

    spec = "I want to eat <food>"
    extras = [
        dragonfly.Choice(
            "food", {"(an | a juicy) apple": "good", "a [greasy] hamburger": "bad"}
        )
    ]

    def _process_recognition(self, node, extras):
        good_or_bad = extras["food"]
        print("That is a %s idea!" % good_or_bad)

class ExampleDictationRule(dragonfly.MappingRule):
    mapping = {
        "dictate <text>": dragonfly.Function(lambda text: print("I heard %r!" % str(text))),
    }
    extras = [ dragonfly.Dictation("text") ]


# Load engine before instantiating rules/grammars!
# Set any configuration options here as keyword arguments.
engine = dragonfly.get_engine("kaldi",
    model_dir='kaldi_model',
    # tmp_dir='kaldi_tmp',  # default for temporary directory
    # vad_aggressiveness=3,  # default aggressiveness of VAD
    # vad_padding_ms=300,  # default ms of required silence surrounding VAD
    # input_device_index=None,  # set to an int to choose a non-default microphone
    # cloud_dictation=None,  # set to 'gcloud' to use cloud dictation
)
# Call connect() now that the engine configuration is set.
engine.connect()

grammar = dragonfly.Grammar(name="mygrammar")
rule = ExampleCustomRule()
grammar.add_rule(rule)
grammar.add_rule(ExampleDictationRule())
grammar.load()

print('Try saying: "I want to eat an apple" or "I want to eat a greasy hamburger" or "dictate this is just a test"')
print("Listening...")
engine.do_recognition()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


"""
SR back-end package for CMU Pocket Sphinx
============================================================================

The main interface to this engine is provided by methods and properties of
the engine class. Please see the :ref:`CMU Pocket Sphinx engine page
<RefSphinxEngine>` for more details.

"""

import logging
_log = logging.getLogger("engine.sphinx")


# Module level singleton instance of this engine implementation.
_engine = None


def is_engine_available():
    """ Check whether the Sphinx engine is available. """
    if _engine:
        return True

    # Attempt to import the engine class from the module.
    try:
        from .engine import SphinxEngine
        return True
    except ImportError as e:
        _log.warning("Failed to import from Sphinx engine module: %s", e)
        return False
    except Exception as e:
        _log.exception("Exception during import of Sphinx engine module: "
                       "%s", e)
        return False

    return True


def get_engine():
    """ Retrieve the Sphinx back-end engine object. """
    global _engine
    if not _engine:
        from .engine import SphinxEngine
        _engine = SphinxEngine()
    return _engine



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements base classes for structured number grammar
elements.

"""

from ..loader             import language
from ...grammar.elements  import Alternative, Repetition, Compound, RuleWrap


#---------------------------------------------------------------------------
# Base class for digit-series element classes.

class Digits(Repetition):

    _content = None
    _digit_name = "_digit"

    def __init__(self, name=None, min=1, max=12, as_int=False, default=None):
        if not self._content:
            self.__class__._content = language.DigitsContent
        self._digits = self._content.digits

        self._as_int = as_int
        if self._as_int: self._base = len(self._digits) - 1

        pairs = []
        for value, word in enumerate(self._digits):
            if isinstance(word, str):
                pairs.append((word, value))
            elif isinstance(word, (tuple, list)):
                pairs.extend([(w, value) for w in word])
            else:
                raise ValueError("Invalid type in digit list: %r" % word)

        alternatives = [Compound(w, value=v, name=self._digit_name)
                        for w, v in pairs]
        child = Alternative(alternatives)
        Repetition.__init__(self, child, min, max, name=name,
                            default=default)

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        arguments = "%d-%d" % (self._min, self._max)
        if self.name is not None:
            arguments = "'%s', %s" % (self.name, arguments)
        return "%s(%s)" % (self.__class__.__name__, arguments)

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def value(self, node):
        children = node.get_children_by_name(self._digit_name)
        digits = [c.value() for c in children]
        if self._as_int:
            value = 0
            for d in digits:
                value *= self._base
                value += d
            return d
        else:
            return digits


#---------------------------------------------------------------------------
# Digits reference class.

class DigitsRef(RuleWrap):

    def __init__(self, name=None, min=1, max=12, as_int=True, default=None):
        element = Digits(name=None, min=min, max=max, as_int=as_int)
        RuleWrap.__init__(self, name, element, default=default)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
FocusWindow action
============================================================================

"""

from .action_base      import ActionBase, ActionError
from ..windows  import Window


#---------------------------------------------------------------------------

class FocusWindow(ActionBase):
    """
        Bring a window to the foreground action.

        Constructor arguments:
         - *executable* (*str*) -- part of the filename of the
           application's executable to which the target window belongs;
           not case sensitive.
         - *title* (*str*) -- part of the title of the target window;
           not case sensitive.
         - *index* (*str* or *int*) -- zero-based index of the target
           window, for multiple matching windows; can be a string (for
           substitution) but must be convertible to an integer.
         - *filter_func* (*callable*) -- called with a single argument
           (the window object), and should return ``True`` for your
           target windows; example:
           ``lambda window: window.get_position().dy > 100``.
         - *focus_only* (*bool*, default *False*) -- if *True*, then
           attempt to focus the window without raising it by using the
           *Window.set_focus()* method instead of *set_foreground()*.
           This argument may do nothing depending on the platform.

        This action searches all visible windows for a window which
        matches the given parameters.

    """

    def __init__(self, executable=None, title=None, index=None,
                 filter_func=None, focus_only=False):
        if executable:  self.executable = executable.lower()
        else:           self.executable = None
        if title:       self.title = title.lower()
        else:           self.title = None
        self.index = index
        self.filter_func = filter_func
        self.focus_only = focus_only
        ActionBase.__init__(self)

        arguments = []
        if executable:  arguments.append("executable=%r" % executable)
        if title:       arguments.append("title=%r" % title)
        if index:       arguments.append("index=%r" % index)
        if filter_func: arguments.append("filter_func=%r" % filter_func)
        if focus_only:  arguments.append("focus_only=%r" % focus_only)
        self._str = ", ".join(arguments)

    def _execute(self, data=None):
        executable = self.executable
        title = self.title
        index = self.index
        if data and isinstance(data, dict):
            if executable:  executable = (executable % data).lower()
            if title:       title = (title % data).lower()
            if index:       index = (index % data).lower()

        index = int(index) if index else 0

        # Get the first matching window and bring it to the foreground.
        windows = Window.get_matching_windows(executable, title)
        if self.filter_func:
            windows = [window for window in windows
                       if self.filter_func(window)]
        if windows and (index < len(windows)):
            window = windows[index]
            if self.focus_only:
                window.set_focus()
            else:
                window.set_foreground()
        else:
            raise ActionError("Failed to find window (%s)." % self._str)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements the compiler base class.
"""


import logging
import dragonfly.grammar.elements as elements_


#---------------------------------------------------------------------------

class CompilerError(Exception):
    pass


#---------------------------------------------------------------------------

class CompilerBase(object):

    _log = logging.getLogger("engine.compiler")

    element_compilers = [
        (elements_.Sequence,    lambda s,e,*a,**k: s._compile_sequence(e,*a,**k)),
        (elements_.Alternative, lambda s,e,*a,**k: s._compile_alternative(e,*a,**k)),
        (elements_.Optional,    lambda s,e,*a,**k: s._compile_optional(e,*a,**k)),
        (elements_.Literal,     lambda s,e,*a,**k: s._compile_literal(e,*a,**k)),
        (elements_.RuleRef,     lambda s,e,*a,**k: s._compile_rule_ref(e,*a,**k)),
        (elements_.ListRef,     lambda s,e,*a,**k: s._compile_list_ref(e,*a,**k)),
        (elements_.Dictation,   lambda s,e,*a,**k: s._compile_dictation(e,*a,**k)),
        (elements_.Impossible,  lambda s,e,*a,**k: s._compile_impossible(e,*a,**k)),
        (elements_.Empty,       lambda s,e,*a,**k: s._compile_empty(e,*a,**k)),
        ]

    #-----------------------------------------------------------------------

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    #-----------------------------------------------------------------------
    # Methods for compiling grammars.

    def compile_grammar(self, grammar, *args, **kwargs):
        raise NotImplementedError("Compiler %s not implemented." % self)

    #-----------------------------------------------------------------------
    # Methods for compiling elements.

    def compile_element(self, element, *args, **kwargs):
        # Look for a compiler method to handle the given element.
        for element_type, compiler in self.element_compilers:
            if isinstance(element, element_type):
                compiler(self, element, *args, **kwargs)
                return

        # Didn't find a compiler method for this element type.
        raise NotImplementedError("Compiler %s not implemented"
                                  " for element type %s."
                                  % (self, element))

    #-----------------------------------------------------------------------

    def _compile_unknown_element(self, element, *args, **kwargs):
        raise NotImplementedError("Compiler %s not implemented"
                                  " for element type %s."
                                  % (self, element))

    _compile_sequence     = _compile_unknown_element
    _compile_alternative  = _compile_unknown_element
    _compile_optional     = _compile_unknown_element
    _compile_literal      = _compile_unknown_element
    _compile_rule_ref     = _compile_unknown_element
    _compile_list_ref     = _compile_unknown_element
    _compile_dictation    = _compile_unknown_element
    _compile_impossible   = _compile_unknown_element
    _compile_empty        = _compile_unknown_element



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file lists all of the basic grammar element classes.

    It is this file which is usually imported by end-user code which
    needs to use dragonfly grammar elements.
"""


import dragonfly.grammar.elements_basic as basic_
import dragonfly.grammar.elements_compound as compound_

#===========================================================================
# Element classes.

#---------------------------------------------------------------------------
# Element base class.

ElementBase     = basic_.ElementBase


#---------------------------------------------------------------------------
# Basic element classes.

Sequence        = basic_.Sequence
Alternative     = basic_.Alternative
Optional        = basic_.Optional
Repetition      = basic_.Repetition
Literal         = basic_.Literal
RuleRef         = basic_.RuleRef
Rule            = basic_.RuleRef        # For backwards compatibility.
ListRef         = basic_.ListRef
List            = basic_.ListRef        # For backwards compatibility.
DictListRef     = basic_.DictListRef
DictList        = basic_.DictListRef    # For backwards compatibility.
Empty           = basic_.Empty
Dictation       = basic_.Dictation
Modifier        = basic_.Modifier
Impossible      = basic_.Impossible
RuleWrap        = basic_.RuleWrap


#---------------------------------------------------------------------------
# Compound element classes.

Compound        = compound_.Compound
Choice          = compound_.Choice



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements the ConnectionGrammar class.
"""

# pylint: disable=expression-not-assigned

from six import string_types

try:
    from win32com.client import Dispatch
    from pywintypes import com_error
except ImportError as error:
    import sys
    if sys.platform.startswith("win"):
        raise error

    # These modules aren't available on non-Windows platforms, so mock what is used.
    class COMError(Exception):
        pass

    com_error = COMError

    class Dispatch(object):
        def __init__(self, _):
            pass

from .grammar_base import Grammar


# ---------------------------------------------------------------------------

class ConnectionGrammar(Grammar):
    """
        Grammar class for maintaining a COM connection well
        within a given context.  This is useful for controlling
        applications through COM while they are in the
        foreground.  This grammar class will take care of
        dispatching the correct COM interface when the
        application comes to the foreground, and releasing it
        when the application is no longer there.

         * ``name`` -- name of this grammar.
         * ``description`` -- description for this grammar.
         * ``context`` -- context within which to maintain
           the COM connection.
         * ``app_name`` -- COM name to dispatch.
    """

    def __init__(self, name, description=None, context=None, app_name=None):
        assert isinstance(app_name, string_types) or app_name is None
        self._app_name = app_name
        self._application = None
        Grammar.__init__(self, name=name, description=description,
                         context=context)

    def __del__(self):
        try:
            self.disconnect()
        except Exception as e:
            self._log.warning("Grammar %s: failed to disconnect from "
                              "%r: %s.", self, self._app_name, e)

    # -----------------------------------------------------------------------
    # Methods for context management.

    application = property(lambda self: self._application,
                           doc="COM handle to the application.")

    def enter_context(self):
        if self.connect():
            self.connection_up()
            return True
        else:
            return False

    def exit_context(self):
        [r.deactivate() for r in self._rules if r.active]
        self.disconnect()
        self.connection_down()

    def _process_begin(self, executable, title, handle):
        # If not connected yet, retry.  If the connection fails after
        #  single attempt, give up.
        if not self._application:
            if not self.connect():
                return False
            self.connection_up()
        return True

    # -----------------------------------------------------------------------
    # Methods for managing the application connection.

    def connect(self):
        if not self._app_name:
            return True
        try:
            self._application = Dispatch(self._app_name)
        except com_error as e:
            if self._log_begin:
                self._log_begin.warning("Grammar %s: failed to connect to "
                                        "%r: %s.", self, self._app_name, e)
            return False
        else:
            [r.activate() for r in self._rules if not r.active]
            return True

    def disconnect(self):
        self._application = None

    def connection_up(self):
        """
            Method called immediately after entering this
            instance's context and successfully setting up its
            connection.

            By default this method doesn't do anything.
            This method should be overridden by derived classes
            if they need to synchronize some internal state with
            the application.  The COM connection is available
            through the ``self.application`` attribute.
        """

    def connection_down(self):
        """
            Method called immediately after exiting this
            instance's context and disconnecting from the
            application.

            By default this method doesn't do anything.
            This method should be overridden by derived classes
            if they need to clean up after disconnection.
        """



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
GrammarWrapper class for the CMU Pocket Sphinx engine
============================================================================

"""


import logging

from jsgf import Literal, filter_expansion

import dragonfly.grammar.state as state_

from ..base import GrammarWrapperBase


class GrammarWrapper(GrammarWrapperBase):

    _log = logging.getLogger("engine")

    def __init__(self, grammar, engine, recobs_manager, search_name):
        """
        :type grammar: Grammar
        :type engine: SphinxEngine
        """
        GrammarWrapperBase.__init__(self, grammar, engine, recobs_manager)
        self.set_search = True
        self._search_name = search_name
        self.exclusive = False

        # Compile the grammar into a JSGF grammar and set the language.
        self._jsgf_grammar = engine.compiler.compile_grammar(grammar)
        self._jsgf_grammar.language_name = engine.language

    def _get_reference_name(self, name):
        return self.engine.compiler.get_reference_name(name)

    def enable_rule(self, name):
        ref_name = self._get_reference_name(name)
        jsgf_rule = self._jsgf_grammar.get_rule_from_name(ref_name)

        # Only enable the rule and set the flag if the rule is disabled.
        if not jsgf_rule.active:
            jsgf_rule.enable()
            self.set_search = True

    def disable_rule(self, name):
        ref_name = self._get_reference_name(name)
        jsgf_rule = self._jsgf_grammar.get_rule_from_name(ref_name)

        # Only disable the rule and set the flag if the rule is enabled.
        if jsgf_rule.active:
            jsgf_rule.disable()
            self.set_search = True

    def update_list(self, lst):
        # Recompile the list again.
        grammar = self._jsgf_grammar
        name = self._get_reference_name(lst.name)
        old_rule = grammar.get_rule_from_name(name)
        new_rule, unknown_words = self.engine.compiler.recompile_list(
            lst, grammar
        )

        # Only replace the old rule if the list has changed.
        if old_rule != new_rule:
            grammar.remove_rule(old_rule, ignore_dependent=True)
            grammar.add_rule(new_rule)
            self.set_search = True

            # Log a warning about unknown words if necessary.
            if unknown_words:
                logger = logging.getLogger("engine.compiler")
                logger.warning("List '%s' used words not found in the "
                               "pronunciation dictionary: %s", name,
                               ", ".join(sorted(unknown_words)))

    def compile_jsgf(self):
        return self._jsgf_grammar.compile_as_root_grammar()

    @property
    def grammar_words(self):
        """
        Set of all words used in this grammar.

        :returns: set
        """
        words = []
        for rule in self._jsgf_grammar.rules:
            rule_literals = filter_expansion(
                rule.expansion, lambda x: isinstance(x, Literal) and x.text,
                shallow=True
            )
            for literal in rule_literals:
                words.extend(literal.text.split())

        # Return a set of words with no duplicates.
        return set(words)

    @property
    def search_name(self):
        """
        The name of the Pocket Sphinx search that the engine should use to
        process speech for the grammar.

        :return: str
        """
        return self._search_name

    @property
    def grammar_active(self):
        """
        Whether the grammar is enabled and has any active rules.
        :rtype: bool
        """
        return self.grammar.enabled and any(self.grammar.active_rules)

    def process_begin(self, executable, title, handle):
        self.grammar.process_begin(executable, title, handle)

    def process_words(self, words):
        # Return early if the grammar is disabled or if there are no active
        # rules.
        if not (self.grammar.enabled and self.grammar.active_rules):
            return

        self._log.debug("Grammar %s: received recognition %r."
                        % (self.grammar.name, words))
        results_obj = None  # TODO Use PS results object once implemented

        # TODO Make special grammar callbacks work properly.
        # These special methods are never called for this engine.
        if words == "other":
            func = getattr(self.grammar, "process_recognition_other", None)
            self._process_grammar_callback(func, words=words,
                                           results=results_obj)
            return
        elif words == "reject":
            func = getattr(self.grammar, "process_recognition_failure",
                           None)
            self._process_grammar_callback(func, results=results_obj)
            return

        # If the words argument was not "other" or "reject", then it is a
        # sequence of (word, rule_id) 2-tuples.
        words_rules = tuple(words)
        words = tuple(word for word, _ in words)

        # Call the grammar's general process_recognition method, if present.
        func = getattr(self.grammar, "process_recognition", None)
        if func:
            if not self._process_grammar_callback(func, words=words,
                                                  results=results_obj):
                # Return early if the method didn't return True or equiv.
                return

        # Iterate through this grammar's rules, attempting to decode each.
        # If successful, call that rule's method for processing the
        # recognition and return.
        s = state_.State(words_rules, self.grammar.rule_names, self.engine)
        for r in self.grammar.rules:
            if not (r.active and r.exported):
                continue
            s.initialize_decoding()
            for _ in r.decode(s):
                if s.finished():
                    # Build the parse tree used to process this rule.
                    root = s.build_parse_tree()

                    # Notify observers using the manager *before*
                    # processing.
                    notify_args = (words, r, root, results_obj)
                    self.recobs_manager.notify_recognition(
                        *notify_args
                    )

                    # Process the rule if not in training mode.
                    if not self.engine.training_session_active:
                        try:
                            r.process_recognition(root)
                            self.recobs_manager.notify_post_recognition(
                                *notify_args
                            )
                        except Exception as e:
                            self._log.exception("Failed to process rule "
                                                "'%s': %s" % (r.name, e))
                    return True

        self._log.debug("Grammar %s: failed to decode recognition %r."
                        % (self.grammar.name, words))
        return False



# This file was part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>

# This keyboard implementation is based on Aenea's x11_libxdo.py file.

import logging
import os
import time

import xdo
import Xlib.display

from ._x11_base import BaseX11Keyboard, KEY_TRANSLATION


class LibxdoKeyboard(BaseX11Keyboard):
    """Static class for typing keys with python-libxdo."""

    _log = logging.getLogger("keyboard")
    display = Xlib.display.Display()
    libxdo = xdo.Xdo(os.environ.get('DISPLAY', ''))

    @classmethod
    def send_keyboard_events(cls, events):
        """
        Send a sequence of keyboard events.

        Positional arguments:
        events -- a sequence of tuples of the form
            (keycode, down, timeout), where
                keycode (str): key symbol.
                down (boolean): True means the key will be pressed down,
                    False means the key will be released.
                timeout (int): number of seconds to sleep after
                    the keyboard event.

        """
        cls._log.debug("Keyboard.send_keyboard_events %r", events)

        # TODO: We can distill this entire loop down to a single libxdo function
        # call when we figure out how to properly user charcode_t entities from
        # libxdo.
        for event in events:
            (key, down, timeout) = event
            delay_micros = int(timeout * 1000.0)
            key = KEY_TRANSLATION.get(key, key)

            # Press/release the key, catching any errors.
            try:
                if down:
                    cls.libxdo.send_keysequence_window_down(0, key, delay_micros)
                else:
                    cls.libxdo.send_keysequence_window_up(0, key, delay_micros)
            except Exception as e:
                cls._log.exception("Failed to type key code %s: %s",
                                   key, e)

            # Sleep after the keyboard event if necessary.
            if timeout:
                time.sleep(timeout)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Dictation container class for Natlink
============================================================================

This class is derived from :class:`DictationContainerBase` and implements
dictation formatting for the Natlink and Dragon NaturallySpeaking engine.

"""

from locale import getpreferredencoding

from six import text_type

from ..base import DictationContainerBase
from .dictation_format import WordFormatter


#---------------------------------------------------------------------------
# Natlink dictation class -- container for a series of dictated words.

class NatlinkDictationContainer(DictationContainerBase):
    """
        Container class for dictated words as recognized by the
        :class:`Dictation` element for the Natlink and DNS engine.

    """

    def __init__(self, words, methods):
        unicode_words = []
        for word in words:
            if isinstance(word, text_type):
                unicode_words.append(word)
            else:
                unicode_words.append(word.decode(getpreferredencoding()))
        DictationContainerBase.__init__(self, words=unicode_words,
                                        methods=methods)

    def format(self):
        """ Format and return this dictation. """
        formatter = WordFormatter()
        formatted = formatter.format_dictation(self._words)
        return self.apply_methods(formatted)
